import json
import os
import shutil
import subprocess
import sys
import time
from pathlib import Path
from typing import Any, NamedTuple, Optional

import pytest
import yaml

from oumi.core.configs import TrainingConfig
from oumi.core.configs.params.training_params import TrainerType
from oumi.utils.io_utils import load_json
from oumi.utils.torch_utils import device_cleanup
from tests import get_configs_dir
from tests.e2e import get_e2e_test_output_dir, is_file_not_empty
from tests.markers import requires_gpus


class TrainTestConfig(NamedTuple):
    test_name: str
    config_path: Path
    max_steps: int
    is_lora: bool = False
    skip: bool = False
    interactive_logs: bool = True

    trainer_type: Optional[TrainerType] = None
    model_max_length: Optional[int] = None
    batch_size: Optional[int] = None
    gradient_accumulation_steps: Optional[int] = None
    dataloader_num_workers: Optional[int] = None
    dataloader_prefetch_factor: Optional[int] = None
    save_steps: Optional[int] = None
    save_final_model: Optional[bool] = None
    enable_wandb: Optional[bool] = False  # Disable `wandb`` by default


def _check_checkpoint_dir(
    dir_path: Path, *, is_lora: bool, validate_extra_files: bool = False
):
    """Helper to verify model directory structure."""
    # Check essential model files
    essential_files = [
        "special_tokens_map.json",
        "tokenizer_config.json",
        "tokenizer.json",
        "trainer_state.json",
        "training_args.bin",
    ]
    if is_lora:
        essential_files = ["adapter_config.json"] + essential_files  # OPE-938
    else:
        essential_files = ["config.json", "generation_config.json"] + essential_files

    for file in essential_files:
        assert (dir_path / file).is_file(), f"Missing {file} in {dir_path}"
        assert is_file_not_empty(dir_path / file), f"Empty {file} in {dir_path}"

    model_basename = "adapter_model" if is_lora else "model"
    model_safetensors = dir_path / f"{model_basename}.safetensors"

    if model_safetensors.exists():
        assert model_safetensors.is_file(), (
            f"Exists but not a file: {model_safetensors}"
        )
        assert is_file_not_empty(model_safetensors), f"Empty {model_safetensors}"
    else:
        # The model is sharded. Let's validate model shards.
        model_index_json = dir_path / f"{model_basename}.safetensors.index.json"
        assert model_index_json.is_file(), (
            f"{model_basename} safetensors missing: "
            f"None of {model_index_json} and {model_safetensors} exists"
        )
        model_shards = list(
            sorted(dir_path.glob(f"{model_basename}-*-of-*.safetensors"))
        )
        assert len(model_shards) > 0, (
            f"No '{model_basename}-*-of-*.safetensors' files found under {dir_path}"
        )
        for model_shard in model_shards:
            assert (model_shard).is_file(), f"Missing {model_shard}"
            assert is_file_not_empty(model_shard), f"Empty {model_shard}"
        index_dict: dict[str, Any] = load_json(model_index_json)
        assert "weight_map" in index_dict, f"No `weights_map` in {model_index_json}"
        assert isinstance(index_dict["weight_map"], dict)
        index_shards = {
            (dir_path / shard) for shard in set(index_dict["weight_map"].values())
        }
        assert index_shards == set(model_shards), (
            "Shards defined in model index are inconsistent with shards on file system"
        )

    if is_lora:
        config = load_json(dir_path / "adapter_config.json")
        for key in ("peft_type", "r", "target_modules", "lora_alpha"):
            assert key in config, f"Invalid model config: Missing '{key}'\n{config}"
    else:  # OPE-938
        # Verify config.json is valid JSON
        config = load_json(dir_path / "config.json")
        assert "model_type" in config, f"Invalid model config:\n{config}"

        # Verify generation config
        gen_config = load_json(dir_path / "generation_config.json")
        assert isinstance(gen_config, dict), "Invalid generation config"

    # Verify special tokens map
    with open(dir_path / "special_tokens_map.json") as f:
        tokens_map = json.load(f)
        assert isinstance(tokens_map, dict), "Invalid special tokens map"

    # Verify tokenizer config
    with open(dir_path / "tokenizer_config.json") as f:
        tok_config = json.load(f)
        assert isinstance(tok_config, dict), "Invalid tokenizer config"

    # Verify tokenizer
    with open(dir_path / "tokenizer.json") as f:
        tokenizer = json.load(f)
        assert isinstance(tokenizer, dict), "Invalid tokenizer file"

    # Verify trainer state
    with open(dir_path / "trainer_state.json") as f:
        trainer_state = json.load(f)
        assert "best_model_checkpoint" in trainer_state, "Invalid trainer state"
        assert "log_history" in trainer_state, "Missing training logs in trainer state"

    if validate_extra_files:
        # Additional checkpoint-specific files
        checkpoint_files = ["scheduler.pt"]
        for file in checkpoint_files:
            assert (dir_path / file).exists(), f"Missing {file} in checkpoint"
            assert is_file_not_empty(dir_path / file), f"Empty {file} in checkpoint"

        optimizer_files = ["optimizer.pt", "optimizer.bin"]
        num_valid_optimizer_files = 0
        for file in optimizer_files:
            optimizer_file = dir_path / file
            if optimizer_file.exists():
                assert is_file_not_empty(optimizer_file), (
                    f"Empty {optimizer_file} in checkpoint"
                )
                num_valid_optimizer_files += 1
        assert num_valid_optimizer_files == 1, (
            f"Exactly one of {optimizer_files} must exist. "
            f"Got: {num_valid_optimizer_files}"
        )

        if (dir_path / "rng_state.pth").exists():
            assert is_file_not_empty(dir_path / "rng_state.pth")
        else:
            rng_state_shards = list(sorted(dir_path.glob("rng_state_*.pth")))
            assert len(rng_state_shards) > 1
            for file in rng_state_shards:
                assert is_file_not_empty(dir_path / file), f"Empty {file} in checkpoint"


def get_train_test_id_fn(val):
    assert isinstance(val, TrainTestConfig), f"{type(val)}: {val}"
    return val.test_name


def _test_train_impl(
    test_config: TrainTestConfig,
    tmp_path: Path,
    *,
    use_distributed: bool,
    cleanup_output_dir_on_success: bool = True,
    telemetry_callback_enabled: bool = True,
):
    device_cleanup()
    if test_config.skip:
        pytest.skip(f"Skipped the test '{test_config.test_name}'!")
        return

    interactive_logs = test_config.interactive_logs

    test_tag = f"[{test_config.test_name}]"

    _START_TIME = time.perf_counter()
    output_dir = get_e2e_test_output_dir(test_config.test_name, tmp_path=tmp_path)
    output_dir.mkdir(parents=True, exist_ok=True)

    try:
        # Copy config file to output directory
        assert test_config.config_path.exists(), (
            f"{test_tag} Path doesn't exist: {test_config.config_path}"
        )
        assert test_config.config_path.is_file(), (
            f"{test_tag} Path is not a file: {test_config.config_path}"
        )

        # Verify the config is loadable
        try:
            TrainingConfig.from_yaml(test_config.config_path)
        except Exception as e:
            raise RuntimeError(
                f"Failed to load training config from: {test_config.config_path}"
            ) from e

        assert test_config.max_steps > 0, f"max_steps: {test_config.max_steps}"

        cmd: list[str] = []
        if use_distributed:
            cmd.append("oumi distributed torchrun -m oumi train")
        else:
            cmd.append("oumi train")

        # Execute training command
        cmd.extend(
            [
                "-c",
                str(test_config.config_path),
                "--training.max_steps",
                str(test_config.max_steps),
                "--training.output_dir",
                str(output_dir / "train"),
                "--training.run_name",
                test_config.test_name,
            ]
        )

        for param_name, param_value in [
            ("model_max_length", test_config.model_max_length),
        ]:
            if param_value is not None:
                cmd.append(f"--model.{param_name}={str(param_value)}")

        for param_name, param_value in [
            ("trainer_type", test_config.trainer_type),
            ("per_device_train_batch_size", test_config.batch_size),
            ("gradient_accumulation_steps", test_config.gradient_accumulation_steps),
            ("dataloader_num_workers", test_config.dataloader_num_workers),
            ("dataloader_prefetch_factor", test_config.dataloader_prefetch_factor),
            ("save_steps", test_config.save_steps),
            ("save_final_model", test_config.save_final_model),
            ("enable_wandb", test_config.enable_wandb),
        ]:
            if param_value is not None:
                cmd.append(f"--training.{param_name}={str(param_value)}")

        env_vars = dict(os.environ)
        if "TOKENIZERS_PARALLELISM" not in env_vars:
            # Resolves the warning: "Avoid using `tokenizers` before the fork ..."
            env_vars["TOKENIZERS_PARALLELISM"] = "false"

        shell_command = " ".join(cmd)
        print(f"{test_tag} Running the command:\n{shell_command}\n")
        device_cleanup()
        result = subprocess.run(
            shell_command,
            shell=True,
            text=True,
            capture_output=(not interactive_logs),
            stdout=(sys.stdout if interactive_logs else None),
            stderr=(sys.stderr if interactive_logs else None),
            env=env_vars,
        )
        duration_sec = time.perf_counter() - _START_TIME
        if result.returncode == 0:
            print(
                f"{test_tag} Training job successfully finished in {duration_sec:.2f}s!"
            )
        else:
            print(
                f"{test_tag} Training job failed with error code: {result.returncode} "
                f"in {duration_sec:.2f}s!"
            )
            if not interactive_logs:
                print(f"{test_tag} STDOUT:\n\n{result.stdout}\n\n")
                print(f"{test_tag} STDERR:\n\n{result.stderr}\n\n")
            assert result.returncode == 0, (
                f"{test_tag} Training failed with error code: {result.returncode}"
                + ("" if interactive_logs else f"\nSTDERR:\n\n{result.stderr}\n")
            )

        # Check output directory exists
        train_output_dir = output_dir / "train"
        assert train_output_dir.exists(), f"{test_tag} Output directory was not created"
        assert train_output_dir.is_dir(), (
            f"{test_tag} Output directory is not a directory"
        )

        # If saving is disabled, then return early.
        if (test_config.save_steps is not None and test_config.save_steps <= 0) and (
            test_config.save_final_model is not None
            and not test_config.save_final_model
        ):
            return

        # Check main output directory structure
        _check_checkpoint_dir(train_output_dir, is_lora=test_config.is_lora)

        # Verify checkpoint directory
        checkpoints = list(train_output_dir.glob("checkpoint-*"))
        assert len(checkpoints) > 0, f"{test_tag} No checkpoints found"

        for checkpoint in checkpoints:
            _check_checkpoint_dir(
                checkpoint, is_lora=test_config.is_lora, validate_extra_files=True
            )

        # Check logs directory
        logs_dir = train_output_dir / "logs"
        assert logs_dir.exists(), f"{test_tag} Logs directory not found"
        rank_logs = list(logs_dir.glob("rank_*.log"))
        num_ranks = len(rank_logs)
        assert num_ranks > 0, f"{test_tag} No rank logs found"
        for idx in range(num_ranks):
            assert is_file_not_empty(rank_logs[idx]), (
                f"{test_tag} Empty rank log file: {rank_logs[idx]}"
            )

        # Check telemetry directory
        telemetry_dir = train_output_dir / "telemetry"
        assert telemetry_dir.exists(), f"{test_tag} Telemetry directory not found"
        assert telemetry_dir.is_dir(), (
            f"{test_tag} Telemetry directory  is not a directory"
        )

        telemetry_files = [
            "devices_info.txt",
            "training_config.yaml",
            "world_size.json",
        ] + (
            [
                "telemetry_callback_metrics_rank0000.json",
                "telemetry_callback_rank0000.json",
            ]
            if telemetry_callback_enabled
            else []
        )

        for file in telemetry_files:
            file_path = telemetry_dir / file
            assert file_path.exists(), f"Missing telemetry file: {file}"
            assert is_file_not_empty(file_path), f"Empty telemetry file: {file}"

        # Verify telemetry content
        with open(telemetry_dir / "training_config.yaml") as f:
            training_config = yaml.safe_load(f)
            assert "model" in training_config, (
                f"{test_tag} Invalid training config: {training_config}"
            )
            assert "training" in training_config, (
                f"{test_tag} Invalid training config: {training_config}"
            )

        with open(telemetry_dir / "world_size.json") as f:
            world_size = json.load(f)
            assert "WORLD_SIZE" in world_size
            assert world_size.get("WORLD_SIZE", None) == num_ranks, (
                f"{test_tag} World size is inconsistent with: {num_ranks}"
            )

    except Exception as e:
        duration_sec = time.perf_counter() - _START_TIME
        print(f"{test_tag} Test failed: {str(e)}")
        print(f"{test_tag} Test duration: {duration_sec:.2f}s")
        print(f"{test_tag} Test artifacts can be found in: {output_dir}")
        raise

    if cleanup_output_dir_on_success:
        # Clean-up temp data to stay under disk quota.
        print(f"{test_tag} Cleaning up output dir on success: '{output_dir}'...")
        shutil.rmtree(output_dir)


@requires_gpus(count=1, min_gb=24.0)
@pytest.mark.parametrize(
    "test_config",
    [
        TrainTestConfig(
            test_name="train_text_llama_1b",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "llama3_2"
                / "sft"
                / "1b_full"
                / "train.yaml"
            ),
            max_steps=3,
            model_max_length=128,
        ),
        TrainTestConfig(
            test_name="pretrain_text_fineweb",
            config_path=(
                get_configs_dir()
                / "examples"
                / "fineweb_ablation_pretraining"
                / "ddp"
                / "train.yaml"
            ),
            batch_size=2,
            max_steps=3,
            model_max_length=512,
        ),
        TrainTestConfig(
            test_name="pretrain_text_gpt2",
            config_path=(
                get_configs_dir() / "recipes" / "gpt2" / "pretraining" / "train.yaml"
            ),
            batch_size=16,
            dataloader_num_workers=2,
            dataloader_prefetch_factor=4,
            max_steps=3,
        ),
        TrainTestConfig(
            test_name="train_text_smollm_135m_sft",
            config_path=(
                get_configs_dir() / "recipes" / "smollm" / "sft" / "135m" / "train.yaml"
            ),
            max_steps=3,
        ),
    ],
    ids=get_train_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.single_gpu
def test_train_text_1gpu_24gb(
    test_config: TrainTestConfig,
    tmp_path: Path,
):
    _test_train_impl(test_config=test_config, tmp_path=tmp_path, use_distributed=False)


@requires_gpus(count=4, min_gb=39.0)
@pytest.mark.parametrize(
    "test_config",
    [
        TrainTestConfig(
            test_name="train_text_qwen3_30b_a3b_trl_sft_lora",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "qwen3"
                / "sft"
                / "30b_a3b_lora"
                / "train.yaml"
            ),
            trainer_type=TrainerType.TRL_SFT,
            max_steps=3,
            save_steps=3,
            is_lora=True,
        ),
        TrainTestConfig(
            test_name="train_text_llama3_1_8b_trl_sft_full",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "llama3_1"
                / "sft"
                / "8b_full"
                / "train.yaml"
            ),
            trainer_type=TrainerType.TRL_SFT,
            max_steps=3,
            save_steps=3,
        ),
    ],
    ids=get_train_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.multi_gpu
def test_train_text_4gpu_40gb(test_config: TrainTestConfig, tmp_path: Path):
    _test_train_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        use_distributed=True,
    )


@requires_gpus(count=4, min_gb=39.0)
@pytest.mark.parametrize(
    "test_config",
    [
        TrainTestConfig(
            test_name="train_mm_qwen2_vl_2b_trl_sft_fft",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "qwen2_vl_2b"
                / "sft"
                / "full"
                / "train.yaml"
            ),
            trainer_type=TrainerType.TRL_SFT,
            max_steps=3,
            save_steps=3,
        ),
        TrainTestConfig(
            test_name="train_mm_qwen2_vl_2b_oumi_fft",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "qwen2_vl_2b"
                / "sft"
                / "full"
                / "train.yaml"
            ),
            trainer_type=TrainerType.OUMI,
            max_steps=3,
            save_steps=0,
            save_final_model=False,
        ),
    ],
    ids=get_train_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.multi_gpu
def test_train_multimodal_4gpu_40gb(test_config: TrainTestConfig, tmp_path: Path):
    _test_train_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        use_distributed=True,
    )


@requires_gpus(count=1, min_gb=39.0)
@pytest.mark.parametrize(
    "test_config",
    [
        TrainTestConfig(
            test_name="train_mm_qwen2_vl_2b_trl_sft_lora",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "qwen2_vl_2b"
                / "sft"
                / "lora"
                / "train.yaml"
            ),
            trainer_type=TrainerType.TRL_SFT,
            max_steps=3,
            save_steps=3,
            is_lora=True,
        ),
    ],
    ids=get_train_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.single_gpu
def test_train_multimodal_lora_1gpu_40gb(test_config: TrainTestConfig, tmp_path: Path):
    _test_train_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        use_distributed=False,
    )


@requires_gpus(count=4, min_gb=79.0)
@pytest.mark.parametrize(
    "test_config",
    [
        TrainTestConfig(
            test_name="train_mm_llama3_2_vision_11b_full",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "llama3_2_vision"
                / "sft"
                / "11b_full"
                / "train.yaml"
            ),
            max_steps=3,
            save_steps=3,
        ),
        TrainTestConfig(
            test_name="train_mm_llama3_2_vision_11b_lora",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "llama3_2_vision"
                / "sft"
                / "11b_lora"
                / "train.yaml"
            ),
            is_lora=True,
            max_steps=3,
            save_steps=3,
        ),
        TrainTestConfig(
            test_name="train_mm_llava_7b_sft_full",
            config_path=(
                get_configs_dir()
                / "recipes"
                / "vision"
                / "llava_7b"
                / "sft"
                / "train.yaml"
            ),
            max_steps=3,
            save_steps=3,
        ),
    ],
    ids=get_train_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.multi_gpu
def test_train_multimodal_fsdp_4gpu_80gb(test_config: TrainTestConfig, tmp_path: Path):
    _test_train_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        use_distributed=True,
    )


@requires_gpus(count=4, min_gb=79.0)
@pytest.mark.parametrize(
    "test_config",
    [
        TrainTestConfig(
            test_name="train_grpo_letter_counting",
            config_path=(
                get_configs_dir()
                / "examples"
                / "letter_counting"
                / "grpo"
                / "train.yaml"
            ),
            max_steps=3,
            save_steps=3,
        ),
    ],
    ids=get_train_test_id_fn,
)
@pytest.mark.e2e
@pytest.mark.multi_gpu
def test_train_grpo_4gpu_80gb(test_config: TrainTestConfig, tmp_path: Path):
    _test_train_impl(
        test_config=test_config,
        tmp_path=tmp_path,
        use_distributed=True,
        telemetry_callback_enabled=False,
    )



import logging
import tempfile
from pathlib import Path
from unittest.mock import call, patch

import pytest
import typer
from typer.testing import CliRunner

import oumi
import oumi.cli.alias
from oumi.cli.cli_utils import CONTEXT_ALLOW_EXTRA_ARGS
from oumi.cli.evaluate import evaluate
from oumi.core.configs import (
    EvaluationConfig,
    EvaluationTaskParams,
    ModelParams,
)
from oumi.utils.logging import logger

runner = CliRunner()


@pytest.fixture
def mock_fetch():
    with patch("oumi.cli.cli_utils.resolve_and_fetch_config") as m_fetch:
        yield m_fetch


@pytest.fixture
def mock_alias():
    with patch("oumi.cli.evaluate.try_get_config_name_for_alias") as try_alias:
        yield try_alias


def _create_eval_config() -> EvaluationConfig:
    return EvaluationConfig(
        output_dir="output/dir",
        tasks=[
            EvaluationTaskParams(
                evaluation_backend="lm_harness",
                task_name="mmlu",
                num_samples=4,
            )
        ],
        model=ModelParams(
            model_name="MlpEncoder",
            trust_remote_code=True,
            tokenizer_name="gpt2",
        ),
    )


#
# Fixtures
#
@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(evaluate)
    yield fake_app


@pytest.fixture
def mock_evaluate():
    with patch.object(oumi, "evaluate", autospec=True) as m_evaluate:
        yield m_evaluate


def test_evaluate_runs(app, mock_evaluate):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "eval.yaml")
        config: EvaluationConfig = _create_eval_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["--config", yaml_path])
        mock_evaluate.assert_has_calls([call(config)])


def test_evaluate_calls_alias(app, mock_evaluate, mock_alias):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "eval.yaml")
        mock_alias.return_value = yaml_path
        config: EvaluationConfig = _create_eval_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["--config", "an_alias"])
        mock_alias.assert_has_calls(
            [
                call("an_alias", oumi.cli.alias.AliasType.EVAL),
            ]
        )
        mock_evaluate.assert_has_calls([call(config)])


def test_evaluate_unparsable_metrics(app, mock_evaluate):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        mock_evaluate.return_value = [
            {
                "results": {
                    "mmlu_college_computer_science": {
                        "alias": "college_computer_science",
                        "acc,none": 0.24,
                        "acc_stderr,none": 0.042923469599092816,
                    }
                }
            },
            {
                "results": {
                    "leaderboard_bbh": {" ": " ", "alias": "leaderboard_bbh"},
                    "leaderboard_bbh_boolean_expressions": {
                        "alias": " - leaderboard_bbh_boolean_expressions",
                        "acc_norm,none": 0.436,
                        "acc_norm_stderr,none": 0.03142556706028128,
                    },
                    "leaderboard_bbh_causal_judgement": {
                        "alias": " - leaderboard_bbh_causal_judgement",
                        "acc_norm,none": 0.5187165775401069,
                        "acc_norm_stderr,none": 0.03663608375537842,
                    },
                    "leaderboard_bbh_date_understanding": {
                        "alias": " - leaderboard_bbh_date_understanding",
                        "acc_norm,none": 0.192,
                        "acc_norm_stderr,none": 0.024960691989171994,
                    },
                    "leaderboard_bbh_disambiguation_qa": {
                        "alias": " - leaderboard_bbh_disambiguation_qa",
                        "acc_norm,none": 0.312,
                        "acc_norm_stderr,none": 0.029361067575219817,
                    },
                    "leaderboard_bbh_formal_fallacies": {
                        "alias": " - leaderboard_bbh_formal_fallacies",
                        "acc_norm,none": 0.468,
                        "acc_norm_stderr,none": 0.031621252575725504,
                    },
                    "leaderboard_bbh_geometric_shapes": {
                        "alias": " - leaderboard_bbh_geometric_shapes",
                        "acc_norm,none": 0.0,
                        "acc_norm_stderr,none": 0.0,
                    },
                    "leaderboard_bbh_hyperbaton": {
                        "alias": " - leaderboard_bbh_hyperbaton",
                        "acc_norm,none": 0.528,
                        "acc_norm_stderr,none": 0.0316364895315444,
                    },
                    "leaderboard_bbh_logical_deduction_five_objects": {
                        "alias": " - leaderboard_bbh_logical_deduction_five_objects",
                        "acc_norm,none": 0.188,
                        "acc_norm_stderr,none": 0.024760377727750513,
                    },
                    "leaderboard_bbh_logical_deduction_seven_objects": {
                        "alias": " - leaderboard_bbh_logical_deduction_seven_objects",
                        "acc_norm,none": 0.14,
                        "acc_norm_stderr,none": 0.02198940964524027,
                    },
                    "leaderboard_bbh_logical_deduction_three_objects": {
                        "alias": " - leaderboard_bbh_logical_deduction_three_objects",
                        "acc_norm,none": 0.34,
                        "acc_norm_stderr,none": 0.030020073605457904,
                    },
                    "leaderboard_bbh_movie_recommendation": {
                        "alias": " - leaderboard_bbh_movie_recommendation",
                        "acc_norm,none": 0.22,
                        "acc_norm_stderr,none": 0.02625179282460584,
                    },
                    "leaderboard_bbh_navigate": {
                        "alias": " - leaderboard_bbh_navigate",
                        "acc_norm,none": 0.42,
                        "acc_norm_stderr,none": 0.03127799950463662,
                    },
                    "leaderboard_bbh_object_counting": {
                        "alias": " - leaderboard_bbh_object_counting",
                        "acc_norm,none": 0.048,
                        "acc_norm_stderr,none": 0.013546884228085693,
                    },
                    "leaderboard_bbh_penguins_in_a_table": {
                        "alias": " - leaderboard_bbh_penguins_in_a_table",
                        "acc_norm,none": 0.2465753424657534,
                        "acc_norm_stderr,none": 0.03579404139909799,
                    },
                    "leaderboard_bbh_reasoning_about_colored_objects": {
                        "alias": " - leaderboard_bbh_reasoning_about_colored_objects",
                        "acc_norm,none": 0.156,
                        "acc_norm_stderr,none": 0.022995023034068734,
                    },
                    "leaderboard_bbh_ruin_names": {
                        "alias": " - leaderboard_bbh_ruin_names",
                        "acc_norm,none": 0.292,
                        "acc_norm_stderr,none": 0.028814320402205638,
                    },
                    "leaderboard_bbh_salient_translation_error_detection": {
                        "alias": " - leaderboard_bbh_salient_translation_error_detection",  # noqa E501
                        "acc_norm,none": 0.14,
                        "acc_norm_stderr,none": 0.02198940964524027,
                    },
                    "leaderboard_bbh_snarks": {
                        "alias": " - leaderboard_bbh_snarks",
                        "acc_norm,none": 0.46629213483146065,
                        "acc_norm_stderr,none": 0.037496800603689866,
                    },
                    "leaderboard_bbh_sports_understanding": {
                        "alias": " - leaderboard_bbh_sports_understanding",
                        "acc_norm,none": 0.46,
                        "acc_norm_stderr,none": 0.031584653891499,
                    },
                    "leaderboard_bbh_temporal_sequences": {
                        "alias": " - leaderboard_bbh_temporal_sequences",
                        "acc_norm,none": 0.284,
                        "acc_norm_stderr,none": 0.02857695873043742,
                    },
                    "leaderboard_bbh_tracking_shuffled_objects_five_objects": {
                        "alias": " - leaderboard_bbh_tracking_shuffled_objects_five_objects",  # noqa E501
                        "acc_norm,none": 0.2,
                        "acc_norm_stderr,none": 0.025348970020979085,
                    },
                    "leaderboard_bbh_tracking_shuffled_objects_seven_objects": {
                        "alias": " - leaderboard_bbh_tracking_shuffled_objects_seven_objects",  # noqa E501
                        "acc_norm,none": 0.12,
                        "acc_norm_stderr,none": 0.020593600596839946,
                    },
                    "leaderboard_bbh_tracking_shuffled_objects_three_objects": {
                        "alias": " - leaderboard_bbh_tracking_shuffled_objects_three_objects",  # noqa E501
                        "acc_norm,none": 0.304,
                        "acc_norm_stderr,none": 0.029150213374159677,
                    },
                    "leaderboard_bbh_web_of_lies": {
                        "alias": " - leaderboard_bbh_web_of_lies",
                        "acc_norm,none": 0.488,
                        "acc_norm_stderr,none": 0.03167708558254708,
                    },
                },
                "start_time": "20250316_150150",
                "duration_sec": 394,
            },
            {"results": {"mmlu": {"parsable_metric": 1.0}}},
            {
                "results": {
                    "q": {
                        "name": "our name",
                        "second_metric": 0.5,
                        "parsable_metric": 0.77,
                    },
                    "unparsable": "real name",
                }
            },
            {"results": []},
        ]
        yaml_path = str(Path(output_temp_dir) / "eval.yaml")
        config: EvaluationConfig = _create_eval_config()
        config.to_yaml(yaml_path)
        result = runner.invoke(app, ["--config", yaml_path])
        mock_evaluate.assert_has_calls([call(config)])
        assert result.exit_code == 0
        tables = [
            """
┏━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━┓
┃ Benchmark                ┃ Metric ┃ Score  ┃ Std Error ┃
┡━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━┩
│ college_computer_science │ Acc    │ 24.00% │ ±4.29%    │
└──────────────────────────┴────────┴────────┴───────────┘""",
            """
┏━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━┳━━━━━━━━━━━┓
┃ Benchmark                                    ┃ Metric   ┃ Score  ┃ Std Error ┃
┡━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━╇━━━━━━━━━━━┩
│  - leaderboard_bbh_boolean_expressions       │ Acc Norm │ 43.60% │ ±3.14%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_causal_judgement          │ Acc Norm │ 51.87% │ ±3.66%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_date_understanding        │ Acc Norm │ 19.20% │ ±2.50%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_disambiguation_qa         │ Acc Norm │ 31.20% │ ±2.94%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_formal_fallacies          │ Acc Norm │ 46.80% │ ±3.16%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_geometric_shapes          │ Acc Norm │ 0.00%  │ ±0.00%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_hyperbaton                │ Acc Norm │ 52.80% │ ±3.16%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 18.80% │ ±2.48%    │
│ leaderboard_bbh_logical_deduction_five_obje… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 14.00% │ ±2.20%    │
│ leaderboard_bbh_logical_deduction_seven_obj… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 34.00% │ ±3.00%    │
│ leaderboard_bbh_logical_deduction_three_obj… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_movie_recommendation      │ Acc Norm │ 22.00% │ ±2.63%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_navigate                  │ Acc Norm │ 42.00% │ ±3.13%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_object_counting           │ Acc Norm │ 4.80%  │ ±1.35%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_penguins_in_a_table       │ Acc Norm │ 24.66% │ ±3.58%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 15.60% │ ±2.30%    │
│ leaderboard_bbh_reasoning_about_colored_obj… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_ruin_names                │ Acc Norm │ 29.20% │ ±2.88%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 14.00% │ ±2.20%    │
│ leaderboard_bbh_salient_translation_error_d… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_snarks                    │ Acc Norm │ 46.63% │ ±3.75%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_sports_understanding      │ Acc Norm │ 46.00% │ ±3.16%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_temporal_sequences        │ Acc Norm │ 28.40% │ ±2.86%    │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 20.00% │ ±2.53%    │
│ leaderboard_bbh_tracking_shuffled_objects_f… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 12.00% │ ±2.06%    │
│ leaderboard_bbh_tracking_shuffled_objects_s… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  -                                           │ Acc Norm │ 30.40% │ ±2.92%    │
│ leaderboard_bbh_tracking_shuffled_objects_t… │          │        │           │
├──────────────────────────────────────────────┼──────────┼────────┼───────────┤
│  - leaderboard_bbh_web_of_lies               │ Acc Norm │ 48.80% │ ±3.17%    │
└──────────────────────────────────────────────┴──────────┴────────┴───────────┘""",
            """
┏━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━┳━━━━━━━━━┳━━━━━━━━━━━┓
┃ Benchmark ┃ Metric          ┃ Score   ┃ Std Error ┃
┡━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━╇━━━━━━━━━╇━━━━━━━━━━━┩
│ mmlu      │ Parsable Metric │ 100.00% │ -         │
└───────────┴─────────────────┴─────────┴───────────┘""",
            """
┏━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━┓
┃ Benchmark  ┃ Metric          ┃ Score     ┃ Std Error ┃
┡━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━┩
│ q          │ Second Metric   │ 50.00%    │ -         │
├────────────┼─────────────────┼───────────┼───────────┤
│ q          │ Parsable Metric │ 77.00%    │ -         │
├────────────┼─────────────────┼───────────┼───────────┤
│ unparsable │ <unknown>       │ <unknown> │ -         │
└────────────┴─────────────────┴───────────┴───────────┘""",
        ]
        for table in tables:
            assert table in result.stdout


def test_evaluate_with_overrides(app, mock_evaluate):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "eval.yaml")
        config: EvaluationConfig = _create_eval_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(
            app,
            [
                "--config",
                yaml_path,
                "--model.tokenizer_name",
                "new_name",
                "--tasks",
                "[{evaluation_backend: lm_harness, num_samples: 5, task_name: mmlu}]",
            ],
        )
        expected_config = _create_eval_config()
        expected_config.model.tokenizer_name = "new_name"
        if expected_config.tasks:
            if expected_config.tasks[0]:
                expected_config.tasks[0].num_samples = 5
        mock_evaluate.assert_has_calls([call(expected_config)])


def test_evaluate_logging_levels(app, mock_evaluate):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        yaml_path = str(Path(output_temp_dir) / "eval.yaml")
        config: EvaluationConfig = _create_eval_config()
        config.to_yaml(yaml_path)
        _ = runner.invoke(app, ["--config", yaml_path, "--log-level", "DEBUG"])
        assert logger.level == logging.DEBUG
        _ = runner.invoke(app, ["--config", yaml_path, "-log", "WARNING"])
        assert logger.level == logging.WARNING


def test_evaluate_with_oumi_prefix(app, mock_evaluate, mock_fetch):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        output_dir = Path(output_temp_dir)
        yaml_path = "oumi://configs/recipes/smollm/evaluation/135m/eval.yaml"
        expected_path = output_dir / "configs/recipes/smollm/evaluation/135m/eval.yaml"

        config: EvaluationConfig = _create_eval_config()
        expected_path.parent.mkdir(parents=True, exist_ok=True)
        config.to_yaml(expected_path)
        mock_fetch.return_value = expected_path

        with patch.dict("os.environ", {"OUMI_DIR": str(output_dir)}):
            result = runner.invoke(app, ["--config", yaml_path])

        assert result.exit_code == 0
        mock_fetch.assert_called_once_with(yaml_path)
        mock_evaluate.assert_has_calls([call(config)])



import re

import numpy as np
import pytest
import torch

from oumi.utils.torch_utils import (
    ModelParameterCount,
    convert_to_list_of_tensors,
    create_ones_like,
    estimate_sample_dict_size_in_bytes,
    freeze_model_layers,
    get_dtype_size_in_bytes,
    get_first_dim_len,
    get_torch_dtype,
    pad_sequences,
    pad_sequences_left_side,
    pad_sequences_right_side,
    pad_to_max_dim_and_stack,
)


def test_valid_model_parameter_count():
    mpc = ModelParameterCount(
        all_params=1000, trainable_params=800, embedding_params=200
    )
    assert mpc.trainable_params_percent == 80.0
    assert mpc.frozen_params_percent == 20.0


def test_model_parameter_count_zero_params():
    mpc = ModelParameterCount(all_params=0, trainable_params=0, embedding_params=0)
    assert mpc.trainable_params_percent == 0.0
    assert mpc.frozen_params_percent == 100.0


@pytest.mark.parametrize(
    "all_params, trainable_params, embedding_params, error_field",
    [
        (-1, 0, 0, "all_params"),
        (100, -5, 0, "trainable_params"),
        (100, 0, -3, "embedding_params"),
    ],
)
def test_model_parameter_count_negative_values(
    all_params, trainable_params, embedding_params, error_field
):
    with pytest.raises(ValueError, match=rf"`{error_field}`.*must be >= 0\."):
        ModelParameterCount(all_params, trainable_params, embedding_params)


def test_model_parameter_relative_sizes():
    with pytest.raises(
        ValueError,
        match=(
            r"`trainable_params` \(200\) cannot be greater than "
            r"`all_params` \(100\)\."
        ),
    ):
        ModelParameterCount(all_params=100, trainable_params=200, embedding_params=50)

    with pytest.raises(
        ValueError,
        match=(
            r"`embedding_params` \(150\) cannot be greater than "
            r"`all_params` \(100\)\."
        ),
    ):
        ModelParameterCount(all_params=100, trainable_params=80, embedding_params=150)


def test_convert_to_list_of_tensors_empty_list():
    result = convert_to_list_of_tensors([])
    assert isinstance(result, list)
    assert len(result) == 0


def test_convert_to_list_of_tensors_from_lists():
    result = convert_to_list_of_tensors([[1, 2, 3, 4], [5], [6, 7]])
    assert isinstance(result, list)
    assert len(result) == 3

    assert isinstance(result[0], torch.Tensor)
    assert result[0].shape == (4,)
    assert result[0].dtype == torch.int64
    assert np.all(result[0].numpy() == np.asarray([1, 2, 3, 4]))

    assert result[1].shape == (1,)
    assert result[1].dtype == torch.int64
    assert np.all(result[1].numpy() == np.asarray([5]))

    assert result[2].shape == (2,)
    assert result[2].dtype == torch.int64
    assert np.all(result[2].numpy() == np.asarray([6, 7]))


def test_convert_to_list_of_tensors_from_ndarrays():
    result = convert_to_list_of_tensors(
        [np.asarray([1, 2, 3, 4]), np.asarray([5]), np.asarray([6, 7])]
    )
    assert isinstance(result, list)
    assert len(result) == 3

    assert isinstance(result[0], torch.Tensor)
    assert result[0].shape == (4,)
    assert result[0].dtype == torch.int64
    assert np.all(result[0].numpy() == np.asarray([1, 2, 3, 4]))

    assert result[1].shape == (1,)
    assert result[1].dtype == torch.int64
    assert np.all(result[1].numpy() == np.asarray([5]))

    assert result[2].shape == (2,)
    assert result[2].dtype == torch.int64
    assert np.all(result[2].numpy() == np.asarray([6, 7]))


def test_convert_to_list_of_tensors_from_tensors():
    result = convert_to_list_of_tensors(
        [
            torch.from_numpy(np.asarray([1, 2, 3, 4])),
            torch.from_numpy(np.asarray([5])),
            torch.from_numpy(np.asarray([6, 7])),
        ]
    )
    assert isinstance(result, list)
    assert len(result) == 3

    assert isinstance(result[0], torch.Tensor)
    assert result[0].shape == (4,)
    assert result[0].dtype == torch.int64
    assert np.all(result[0].numpy() == np.asarray([1, 2, 3, 4]))

    assert result[1].shape == (1,)
    assert result[1].dtype == torch.int64
    assert np.all(result[1].numpy() == np.asarray([5]))

    assert result[2].shape == (2,)
    assert result[2].dtype == torch.int64
    assert np.all(result[2].numpy() == np.asarray([6, 7]))


def test_pad_sequences_invalid_side():
    test_sequences = [[1, 2, 3, 4], [5], [6, 7]]
    with pytest.raises(ValueError, match="Unsupported padding side: 'bottom'"):
        pad_sequences(test_sequences, padding_side="bottom", padding_value=1)


@pytest.mark.parametrize(
    "padding_value",
    [0, -100, 7],
)
def test_pad_sequences_right_side(padding_value: int):
    test_sequences = [[1, 2, 3, 4], [5], [6, 7]]
    result = pad_sequences_right_side(test_sequences, padding_value=padding_value)
    assert isinstance(result, torch.Tensor)
    assert result.shape == (3, 4)

    pad = padding_value
    assert np.all(
        result.numpy()
        == np.asarray([[1, 2, 3, 4], [5, pad, pad, pad], [6, 7, pad, pad]])
    )

    assert np.all(
        result.numpy()
        == pad_sequences(
            test_sequences, padding_side="right", padding_value=padding_value
        ).numpy()
    )

    # Verify `stack_and_pad_to_max_dim()` returns exact same result for 1D sequences.
    assert np.all(
        result.numpy()
        == pad_to_max_dim_and_stack(
            test_sequences, padding_side="right", padding_value=padding_value
        ).numpy()
    )
    assert np.all(
        result.numpy()
        == pad_to_max_dim_and_stack(test_sequences, padding_value=padding_value).numpy()
    )


@pytest.mark.parametrize(
    "padding_value",
    [0, -100, 7],
)
def test_pad_sequences_left_side(padding_value: int):
    test_sequences = [[1, 2, 3, 4], [5], [6, 7]]
    result = pad_sequences_left_side(test_sequences, padding_value=padding_value)
    assert isinstance(result, torch.Tensor)
    assert result.shape == (3, 4)

    pad = padding_value
    assert np.all(
        result.numpy()
        == np.asarray([[1, 2, 3, 4], [pad, pad, pad, 5], [pad, pad, 6, 7]])
    )

    assert np.all(
        result.numpy()
        == pad_sequences(
            test_sequences, padding_side="left", padding_value=padding_value
        ).numpy()
    )

    # Verify `stack_and_pad_to_max_dim()` returns exact same result for 1D sequences.
    assert np.all(
        result.numpy()
        == pad_to_max_dim_and_stack(
            test_sequences, padding_side="left", padding_value=padding_value
        ).numpy()
    )


def test_stack_and_pad_to_max_dim_invalid_side():
    test_sequences = [[1, 2, 3, 4], [5], [6, 7]]
    with pytest.raises(ValueError, match="Unsupported padding side: 'top'"):
        pad_to_max_dim_and_stack(test_sequences, padding_side="top", padding_value=1)


def test_stack_and_pad_to_max_dim_incompatible_dimensionality():
    test_sequences = [[1, 2, 3, 4], [5], [[6, 7], [8, 9]]]
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Tensors have different number of dimensions: 1 vs 2! "
            "Shapes: torch.Size([4]), torch.Size([2, 2])"
        ),
    ):
        pad_to_max_dim_and_stack(test_sequences, padding_value=1)


@pytest.mark.parametrize(
    "padding_value",
    [0, -100, 7],
)
def test_stack_and_pad_to_max_dim_right_side(padding_value):
    test_sequences = [torch.ones([5, 1, 2]), torch.full([3, 2, 1], 2)]
    result = pad_to_max_dim_and_stack(test_sequences, padding_value=padding_value)
    assert result.shape == (2, 5, 2, 2)

    expected = torch.full((2, 5, 2, 2), padding_value)
    expected[0, :, :1, :] = 1
    expected[1, :3, :, :1] = 2
    assert np.all(result.numpy() == expected.numpy()), (
        f"result: {result} expected: {expected}"
    )

    assert np.all(
        expected.numpy()
        == pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=3,
            padding_side="right",
            padding_value=padding_value,
        ).numpy()
    )

    assert np.all(
        expected.numpy()
        == pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=-1,
            padding_value=padding_value,
        ).numpy()
    )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Too many dimensions with variable size. "
            "Got: 3 variable size dimensions. "
            "Maximum allowed: 1"
        ),
    ):
        pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=1,
            padding_side="right",
            padding_value=padding_value,
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Too many dimensions with variable size. "
            "Got: 3 variable size dimensions. "
            "Maximum allowed: 2"
        ),
    ):
        pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=2,
            padding_value=padding_value,
        )


@pytest.mark.parametrize(
    "padding_value",
    [0, -100, 7],
)
def test_stack_and_pad_to_max_dim_left_side(padding_value):
    test_sequences = [
        torch.ones([5, 1, 2, 1]),
        torch.full([3, 2, 1, 1], 2),
        torch.full([1, 1, 1, 1], -3),
    ]
    result = pad_to_max_dim_and_stack(
        test_sequences,
        padding_side="left",
        padding_value=padding_value,
    )
    assert result.shape == (3, 5, 2, 2, 1)

    expected = torch.full((3, 5, 2, 2, 1), padding_value)
    expected[0, :, -1:, :, :] = 1
    expected[1, -3:, :, -1:, :] = 2
    expected[2, -1:, -1:, -1:, :] = -3
    assert np.all(result.numpy() == expected.numpy()), (
        f"result: {result} expected: {expected}"
    )

    assert np.all(
        expected.numpy()
        == pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=3,
            padding_side="left",
            padding_value=padding_value,
        ).numpy()
    )

    assert np.all(
        expected.numpy()
        == pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=-1,
            padding_side="left",
            padding_value=padding_value,
        ).numpy()
    )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Too many dimensions with variable size. "
            "Got: 3 variable size dimensions. "
            "Maximum allowed: 1"
        ),
    ):
        pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=1,
            padding_side="left",
            padding_value=padding_value,
        )

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Too many dimensions with variable size. "
            "Got: 3 variable size dimensions. "
            "Maximum allowed: 2"
        ),
    ):
        pad_to_max_dim_and_stack(
            test_sequences,
            max_variable_sized_dims=2,
            padding_side="left",
            padding_value=padding_value,
        )


def test_create_ones_from_empty():
    result = create_ones_like([])
    assert isinstance(result, list) and len(result) == 0

    result = create_ones_like(np.asarray([]))
    assert isinstance(result, np.ndarray) and result.shape == (0,)

    result = create_ones_like(torch.Tensor(np.asarray([])))
    assert isinstance(result, torch.Tensor) and result.shape == (0,)


def test_create_ones_from_none():
    with pytest.raises(ValueError, match="Unsupported type"):
        create_ones_like(None)


def test_create_ones_from_primitive():
    with pytest.raises(ValueError, match="Unsupported type"):
        create_ones_like(1)
    with pytest.raises(ValueError, match="Unsupported type"):
        create_ones_like(2.0)
    with pytest.raises(ValueError, match="Unsupported type"):
        create_ones_like("zzz")


def test_create_ones_like_inhomogeneous_shape():
    with pytest.raises(
        ValueError,
        match=(
            "setting an array element with a sequence. "
            "The requested array has an inhomogeneous shape after"
        ),
    ):
        create_ones_like([2, 3, [4, 5]])

    with pytest.raises(
        ValueError,
        match=(
            "setting an array element with a sequence. "
            "The requested array has an inhomogeneous shape after"
        ),
    ):
        create_ones_like([2, 3, np.asarray([4, 5]), 6])

    with pytest.raises(
        ValueError,
        match=(
            "setting an array element with a sequence. "
            "The requested array has an inhomogeneous shape after"
        ),
    ):
        create_ones_like([2, torch.Tensor([4, 5]), 7])


def test_create_ones_like_different_types():
    with pytest.raises(
        ValueError,
        match=("Sequence contains elements of different types"),
    ):
        create_ones_like([[4, 5], 1])

    with pytest.raises(
        ValueError,
        match=("Sequence contains elements of different types"),
    ):
        create_ones_like([[4, 5], np.asarray([6, 7])])

    with pytest.raises(
        ValueError,
        match=("Sequence contains elements of different types"),
    ):
        create_ones_like([np.asarray([6, 7]), torch.Tensor([8, 9])])

    with pytest.raises(
        ValueError,
        match=("Sequence contains elements of different types"),
    ):
        create_ones_like([torch.Tensor([8, 9]), [1, 2]])

    with pytest.raises(
        ValueError,
        match=("Sequence contains elements of different types"),
    ):
        create_ones_like([torch.Tensor([8, 9]), "str"])


def test_create_ones_like_success_list():
    result = create_ones_like([2])
    assert isinstance(result, list)
    assert np.all(np.asarray(result) == [1])

    result = create_ones_like([2, 3, 4])
    assert isinstance(result, list)
    assert np.all(np.asarray(result) == [1, 1, 1])

    result = create_ones_like([[2, 3], [4, 5]])
    assert isinstance(result, list)
    assert np.all(np.asarray(result) == np.asarray([[1, 1], [1, 1]]))


def test_create_ones_like_success_numpy():
    result = create_ones_like(np.asarray([2]))
    assert isinstance(result, np.ndarray)
    assert np.all(result == [1])

    result = create_ones_like(np.asarray([2, 3, 4]))
    assert isinstance(result, np.ndarray)
    assert np.all(np.asarray(result) == [1, 1, 1])

    result = create_ones_like([np.asarray([2, 3]), np.asarray([4, 5, 6])])
    assert isinstance(result, list)
    assert isinstance(result[0], np.ndarray)
    assert np.all(result[0] == np.asarray([1, 1]))
    assert isinstance(result[1], np.ndarray)
    assert np.all(result[1] == np.asarray([1, 1, 1]))


def test_create_ones_like_success_tensor():
    result = create_ones_like(torch.Tensor([2]))
    assert isinstance(result, torch.Tensor)
    assert np.all(result.numpy() == np.asarray([1]))

    result = create_ones_like(torch.Tensor([2, 3, 4]))
    assert isinstance(result, torch.Tensor)
    assert np.all(result.numpy() == np.asarray([1, 1, 1]))

    result = create_ones_like([torch.Tensor([2, 3]), torch.Tensor([4, 5, 6])])
    assert isinstance(result, list)
    assert isinstance(result[0], torch.Tensor)
    assert np.all(result[0].numpy() == np.asarray([1, 1]))
    assert isinstance(result[1], torch.Tensor)
    assert np.all(result[1].numpy() == np.asarray([1, 1, 1]))


@pytest.mark.parametrize(
    "dtype_str, expected_dtype",
    [
        ("f64", torch.float64),
        ("float64", torch.float64),
        ("double", torch.float64),
        ("f32", torch.float32),
        ("float32", torch.float32),
        ("float", torch.float32),
        ("bf16", torch.bfloat16),
        ("bfloat16", torch.bfloat16),
        ("f16", torch.float16),
        ("float16", torch.float16),
        ("half", torch.float16),
        ("uint8", torch.uint8),
    ],
)
def test_get_torch_dtype(dtype_str, expected_dtype):
    result = get_torch_dtype(dtype_str)
    assert result == expected_dtype


def test_get_torch_dtype_invalid():
    with pytest.raises(ValueError, match="Unsupported torch dtype: invalid_dtype"):
        get_torch_dtype("invalid_dtype")


def test_get_first_dim_len_list():
    assert get_first_dim_len([]) == 0
    assert get_first_dim_len([1]) == 1
    assert get_first_dim_len([1, 2, 3]) == 3
    assert get_first_dim_len([1, [[6, 9], 7], "abc"]) == 3
    assert get_first_dim_len([[[6, 9], 7], "abc"]) == 2


def test_get_first_dim_len_numpy_array():
    assert get_first_dim_len(np.asarray([])) == 0
    assert get_first_dim_len(np.asarray([1])) == 1
    assert get_first_dim_len(np.asarray([1, 2, 3])) == 3
    assert get_first_dim_len(np.asarray([[1, 2, 3]])) == 1
    assert get_first_dim_len(np.asarray([[1, 2, 3], [1, 2, 3]])) == 2
    assert get_first_dim_len(np.asarray([["a1", "a2", "a3"], ["x1", "x2", "x3"]])) == 2


def test_get_first_dim_len_torch_tensor():
    assert get_first_dim_len(torch.from_numpy(np.asarray([]))) == 0
    assert get_first_dim_len(torch.from_numpy(np.asarray([1]))) == 1
    assert get_first_dim_len(torch.from_numpy(np.asarray([1, 2, 3]))) == 3
    assert get_first_dim_len(torch.from_numpy(np.asarray([[1, 2, 3]]))) == 1
    assert get_first_dim_len(torch.from_numpy(np.asarray([[1, 2, 3], [1, 2, 3]]))) == 2


def test_get_first_dim_len_bad_input_type():
    with pytest.raises(ValueError, match="Unsupported type"):
        get_first_dim_len(None)
    with pytest.raises(ValueError, match="Unsupported type"):
        get_first_dim_len("hello")
    with pytest.raises(ValueError, match="Unsupported type"):
        get_first_dim_len(123)
    with pytest.raises(ValueError, match="Unsupported type"):
        get_first_dim_len(float(123))
    with pytest.raises(ValueError, match="Unsupported type"):
        get_first_dim_len(test_get_first_dim_len_bad_input_type)


def test_get_dtype_size_in_bytes_str():
    assert get_dtype_size_in_bytes("f64") == 8
    assert get_dtype_size_in_bytes("float64") == 8
    assert get_dtype_size_in_bytes("double") == 8
    assert get_dtype_size_in_bytes("f32") == 4
    assert get_dtype_size_in_bytes("float32") == 4
    assert get_dtype_size_in_bytes("float") == 4
    assert get_dtype_size_in_bytes("bf16") == 2
    assert get_dtype_size_in_bytes("bfloat16") == 2
    assert get_dtype_size_in_bytes("f16") == 2
    assert get_dtype_size_in_bytes("float16") == 2
    assert get_dtype_size_in_bytes("half") == 2
    assert get_dtype_size_in_bytes("uint8") == 1


def test_get_dtype_size_in_bytes_torch():
    assert get_dtype_size_in_bytes(torch.float64) == 8
    assert get_dtype_size_in_bytes(torch.float32) == 4
    assert get_dtype_size_in_bytes(torch.float16) == 2
    assert get_dtype_size_in_bytes(torch.uint8) == 1

    assert get_dtype_size_in_bytes(torch.uint64) == 8
    assert get_dtype_size_in_bytes(torch.int64) == 8
    assert get_dtype_size_in_bytes(torch.uint32) == 4
    assert get_dtype_size_in_bytes(torch.int32) == 4
    assert get_dtype_size_in_bytes(torch.uint16) == 2
    assert get_dtype_size_in_bytes(torch.int16) == 2
    assert get_dtype_size_in_bytes(torch.uint8) == 1
    assert get_dtype_size_in_bytes(torch.int8) == 1


def test_get_dtype_size_in_bytes_numpy():
    assert get_dtype_size_in_bytes(np.float64) == 8
    assert get_dtype_size_in_bytes(torch.float32) == 4
    assert get_dtype_size_in_bytes(torch.float16) == 2
    assert get_dtype_size_in_bytes(torch.uint8) == 1

    assert get_dtype_size_in_bytes(torch.uint64) == 8
    assert get_dtype_size_in_bytes(torch.int64) == 8
    assert get_dtype_size_in_bytes(torch.uint32) == 4
    assert get_dtype_size_in_bytes(torch.int32) == 4
    assert get_dtype_size_in_bytes(torch.uint16) == 2
    assert get_dtype_size_in_bytes(torch.int16) == 2
    assert get_dtype_size_in_bytes(torch.uint8) == 1
    assert get_dtype_size_in_bytes(torch.int8) == 1


def test_estimate_sample_dict_size_in_bytes():
    assert estimate_sample_dict_size_in_bytes({"hi": [1] * 100}) == 402
    assert estimate_sample_dict_size_in_bytes({"hi": [1] * 40, "bye": [1] * 60}) == 405
    assert estimate_sample_dict_size_in_bytes({"hi": "Wir müssen"}) == 13
    assert estimate_sample_dict_size_in_bytes({"hi": "Мы должны"}) == 19
    assert estimate_sample_dict_size_in_bytes({"Мы": "должны"}) == 16
    assert estimate_sample_dict_size_in_bytes({"hi": "there"}) == 7

    assert estimate_sample_dict_size_in_bytes({"hi": ["there"]}) == 7

    # Numpy
    assert (
        estimate_sample_dict_size_in_bytes(
            {"foo": np.asarray([[1, 2, 3], [1, 2, 3]], dtype=np.int64)}
        )
        == 3 + 6 * 8
    )
    assert (
        estimate_sample_dict_size_in_bytes(
            {"foo": np.asarray([[1, 2, 3], [1, 2, 3]], dtype=np.int32)}
        )
        == 3 + 6 * 4
    )

    # Torch
    assert (
        estimate_sample_dict_size_in_bytes(
            {
                "foo": torch.from_numpy(
                    np.asarray([[1, 2, 3], [1, 2, 3]], dtype=np.int64)
                )
            }
        )
        == 3 + 6 * 8
    )
    assert (
        estimate_sample_dict_size_in_bytes(
            {
                "foo": torch.from_numpy(
                    np.asarray([[1, 2, 3], [1, 2, 3]], dtype=np.int32)
                )
            }
        )
        == 3 + 6 * 4
    )

    # Mixed
    assert (
        estimate_sample_dict_size_in_bytes(
            {
                "torch": torch.from_numpy(
                    np.asarray([[1, 2, 3], [1, 2, 3]], dtype=np.int64)
                ),
                "numpy": np.asarray([[1, 2, 3], [1, 2, 3]], dtype=np.int32),
                "list": [[1, 2, 3], [1, 2, 3]],
            }
        )
        == 14 + 6 * 8 + 6 * 4 + 6 * 4
    )


class LittleNeuralNetwork(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.foo_flatten = torch.nn.Flatten()
        self.linear_relu_stack = torch.nn.Sequential()
        self.linear_relu_stack.add_module("linearA", torch.nn.Linear(14, 16))
        self.linear_relu_stack.add_module("reluA", torch.nn.ReLU())
        self.linear_relu_stack.add_module("linearB", torch.nn.Linear(16, 12))
        self.linear_relu_stack.add_module("reluB", torch.nn.ReLU())
        self.linear_relu_stack.add_module("linearC", torch.nn.Linear(12, 10))
        nested_stack = torch.nn.Sequential()
        self.linear_relu_stack.add_module("nested", nested_stack)
        nested_stack.add_module("linearX", torch.nn.Linear(10, 8))
        nested_stack.add_module("linearY", torch.nn.Linear(8, 8))
        nested_stack.add_module("linearZ", torch.nn.Linear(8, 4))
        self.bar_softmax = torch.nn.Softmax()

    def forward(self, x):
        x = self.foo_flatten(x)
        logits = self.linear_relu_stack(x)
        return self.bar_softmax(logits)


def _create_test_neural_network() -> torch.nn.Module:
    return LittleNeuralNetwork().to("cpu")


def test_freeze_model_layers():
    assert freeze_model_layers(_create_test_neural_network(), []) == 0
    assert (
        freeze_model_layers(_create_test_neural_network(), ["linear_relu_stack"]) == 1
    )
    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            ["linear_relu_stack", "foo_flatten", "bar_softmax"],
        )
        == 3
    ), "Basic layers not handled correctly"

    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            [
                "bar_softmax",
                "foo_flatten",
                "linear_relu_stack.linearA",
                "linear_relu_stack.linearB",
            ],
        )
        == 4
    ), "Basic nested layers not handled correctly"

    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            [
                "bar_softmax",
                "bar_softmax",
                "foo_flatten",
                "linear_relu_stack.linearA",
                "linear_relu_stack.linearB",
                "linear_relu_stack.linearB",
            ],
        )
        == 4
    ), "Duplicates not handled correctly"

    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            [
                "bar_softmax",
                "foo_flatten",
                "linear_relu_stack",
                "linear_relu_stack.linearA",  # has no effect as parent is frozen.
                "linear_relu_stack.linearB",  # has no effect as parent is frozen.
            ],
        )
        == 3
    ), "Redundant children not handled correctly"

    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            [
                "bar_softmax",
                "foo_flatten_DOESNT_EXIST",
                "linear_relu_stack.linearA",
                "linear_relu_stack.linearDOESNT_EXIST",
            ],
        )
        == 2
    ), "Non-existent entries not handled correctly"

    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            [
                "bar_softmax",
                "foo_flatten",
                "linear_relu_stack.linearA",
                "linear_relu_stack.linearB",
                "linear_relu_stack.nested.linearX",
                "linear_relu_stack.nested.linearZ",
            ],
        )
        == 6
    ), "Multiple nested layers not handled correctly"

    assert (
        freeze_model_layers(
            _create_test_neural_network(),
            [
                "bar_softmax",
                "foo_flatten",
                "linear_relu_stack.linearA",
                "linear_relu_stack.linearB",
                "linear_relu_stack.nested.linearX",  # no effect as parent is frozen.
                "linear_relu_stack.nested.linearY",  # no effect as parent is frozen.
                "linear_relu_stack.nested.linearZ",  # no effect as parent is frozen.
                "linear_relu_stack.nested",
            ],
        )
        == 5
    ), "Redunant multiple nested layers not handled correctly"



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Annotated, Any, Callable, Optional

import typer
from rich.table import Table

from oumi.cli import cli_utils


def judge_dataset_file(
    ctx: typer.Context,
    judge_config: Annotated[
        str,
        typer.Option(
            "--config",
            help="Path to the judge config file",
        ),
    ],
    input_file: Annotated[
        str, typer.Option("--input", help="Path to the dataset input file (jsonl)")
    ],
    output_file: Annotated[
        Optional[str],
        typer.Option("--output", help="Path to the output file (jsonl)"),
    ] = None,
    display_raw_output: bool = False,
):
    """Judge a dataset."""
    # Delayed import
    from oumi import judge

    judge_file(
        ctx=ctx,
        judge_config=judge_config,
        input_file=input_file,
        output_file=output_file,
        display_raw_output=display_raw_output,
        judgment_fn=judge.judge_dataset_file,
    )


def judge_conversations_file(
    ctx: typer.Context,
    judge_config: Annotated[
        str,
        typer.Option(
            "--config",
            help="Path to the judge config file",
        ),
    ],
    input_file: Annotated[
        str, typer.Option("--input", help="Path to the dataset input file (jsonl)")
    ],
    output_file: Annotated[
        Optional[str],
        typer.Option("--output", help="Path to the output file (jsonl)"),
    ] = None,
    display_raw_output: bool = False,
):
    """Judge a list of conversations."""
    # Delayed import
    from oumi import judge

    judge_file(
        ctx=ctx,
        judge_config=judge_config,
        input_file=input_file,
        output_file=output_file,
        display_raw_output=display_raw_output,
        judgment_fn=judge.judge_conversations_file,
    )


def judge_file(
    ctx: typer.Context,
    judge_config: Annotated[
        str,
        typer.Option(
            "--config",
            help="Path to the judge config file",
        ),
    ],
    input_file: Annotated[
        str, typer.Option("--input", help="Path to the dataset input file (jsonl)")
    ],
    output_file: Annotated[
        Optional[str],
        typer.Option("--output", help="Path to the output file (jsonl)"),
    ] = None,
    display_raw_output: bool = False,
    judgment_fn: Callable[..., list[Any]] = ...,
):
    """Judge a dataset or list of conversations."""
    # Delayed import
    from oumi.core.configs.judge_config import JudgeConfig

    # Load configs
    extra_args = cli_utils.parse_extra_cli_args(ctx)

    # Resolve judge config
    judge_config_obj = JudgeConfig.from_path(path=judge_config, extra_args=extra_args)

    # Ensure the dataset input file exists
    if not Path(input_file).exists():
        typer.echo(f"Input file not found: '{input_file}'")
        raise typer.Exit(code=1)

    # Judge the dataset
    judge_outputs = judgment_fn(
        judge_config=judge_config_obj,
        input_file=input_file,
        output_file=output_file,
    )

    # Calculate the overall score
    overall_score = 0.0
    for judge_output in judge_outputs:
        judgment_score = judge_output.field_scores.get("judgment", None)
        if judgment_score is not None:
            overall_score += judgment_score
        else:
            overall_score = None
            break

    # Display the overall score
    if overall_score is not None:
        overall_score = overall_score / len(judge_outputs)
        cli_utils.CONSOLE.print(
            f"\n[bold blue]Overall Score: {overall_score:.2%}[/bold blue]"
        )

    # Display the judge outputs if no output file was specified
    if not output_file:
        table = Table(
            title="Judge Results",
            title_style="bold magenta",
            show_edge=False,
            show_lines=True,
        )
        table.add_column("Judgment", style="cyan")
        table.add_column("Judgment Score", style="green")
        table.add_column("Explanation", style="yellow")
        if display_raw_output:
            table.add_column("Raw Output", style="white")

        for judge_output in judge_outputs:
            judgment_value = str(judge_output.field_values.get("judgment", "N/A"))
            judgment_score = str(judge_output.field_scores.get("judgment", "N/A"))
            explanation_value = str(judge_output.field_values.get("explanation", "N/A"))

            if display_raw_output:
                table.add_row(
                    judgment_value,
                    judgment_score,
                    explanation_value,
                    judge_output.raw_output,
                )
            else:
                table.add_row(judgment_value, judgment_score, explanation_value)

        cli_utils.CONSOLE.print(table)
    else:
        typer.echo(f"Results saved to {output_file}")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Union

import pandas as pd

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("allenai/tulu-3-sft-mixture")
class Tulu3MixtureDataset(BaseSftDataset):
    default_dataset = "allenai/tulu-3-sft-mixture"

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Convert the example into a Conversation.

        Args:
            example (dict or Pandas Series): An example containing a `messages`
               field which is a list of dicts with `content` and `role` string
               fields
        """
        messages = [self._transform_message(msg) for msg in example["messages"]]
        return Conversation(messages=messages)

    @classmethod
    def _transform_message(cls, message: dict) -> Message:
        roles = {"system": Role.SYSTEM, "assistant": Role.ASSISTANT, "user": Role.USER}
        content = message["content"]
        role_str = message["role"].lower().strip()
        try:
            role = roles[role_str]
            return Message(role=role, content=content)
        except KeyError as e:
            raise ValueError(
                f"Unknown role {message['role']}, was expecting one of {roles.keys()}"
            ) from e



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from oumi.core.configs.params.gkd_params import GkdParams


def test_teacher_model_init_kwargs_uses_dtype_not_torch_dtype():
    """Test that default uses 'dtype' instead of 'torch_dtype'."""
    params = GkdParams()

    # Check default has 'dtype'
    assert "dtype" in params.teacher_model_init_kwargs
    assert params.teacher_model_init_kwargs["dtype"] == "auto"


def test_to_hf_trainer_kwargs_with_default_teacher_init_kwargs():
    """Test to_hf_trainer_kwargs with default empty teacher_model_init_kwargs."""
    params = GkdParams(teacher_model_init_kwargs={})
    kwargs = params.to_hf_trainer_kwargs()

    # Should initialize empty dict and add dtype
    assert "teacher_model_init_kwargs" in kwargs
    assert kwargs["teacher_model_init_kwargs"] == {"dtype": "auto"}


def test_to_hf_trainer_kwargs_preserves_dtype():
    """Test that existing 'dtype' in teacher_model_init_kwargs is preserved."""
    params = GkdParams(
        teacher_model_init_kwargs={
            "dtype": "float16",
            "device_map": "auto",
        }
    )
    kwargs = params.to_hf_trainer_kwargs()

    # Should preserve the custom dtype
    assert kwargs["teacher_model_init_kwargs"]["dtype"] == "float16"
    assert kwargs["teacher_model_init_kwargs"]["device_map"] == "auto"


def test_to_hf_trainer_kwargs_adds_dtype_when_missing():
    """Test that 'dtype' is added when not present in teacher_model_init_kwargs."""
    params = GkdParams(
        teacher_model_init_kwargs={
            "device_map": "auto",
            "attn_implementation": "flash_attention_2",
        }
    )
    kwargs = params.to_hf_trainer_kwargs()

    # Should add dtype = "auto" when not present
    assert kwargs["teacher_model_init_kwargs"]["dtype"] == "auto"
    assert kwargs["teacher_model_init_kwargs"]["device_map"] == "auto"
    assert (
        kwargs["teacher_model_init_kwargs"]["attn_implementation"]
        == "flash_attention_2"
    )


def test_to_hf_trainer_kwargs_with_nonempty_teacher_init_kwargs():
    """Test to_hf_trainer_kwargs when teacher_model_init_kwargs is not empty."""
    params = GkdParams(
        teacher_model_init_kwargs={
            "device_map": "cuda:0",
        }
    )
    kwargs = params.to_hf_trainer_kwargs()

    assert "teacher_model_init_kwargs" in kwargs
    assert kwargs["teacher_model_init_kwargs"]["device_map"] == "cuda:0"
    # Should add dtype since it's missing
    assert kwargs["teacher_model_init_kwargs"]["dtype"] == "auto"


def test_to_hf_trainer_kwargs_includes_all_fields():
    """Test that to_hf_trainer_kwargs includes all expected fields."""
    params = GkdParams(
        temperature=0.8,
        lmbda=0.6,
        beta=0.7,
        max_new_tokens=256,
        disable_dropout=False,
        seq_kd=True,
    )
    kwargs = params.to_hf_trainer_kwargs()

    assert kwargs["temperature"] == 0.8
    assert kwargs["lmbda"] == 0.6
    assert kwargs["beta"] == 0.7
    assert kwargs["max_new_tokens"] == 256
    assert kwargs["disable_dropout"] is False
    assert kwargs["seq_kd"] is True
    assert "teacher_model_init_kwargs" in kwargs



from oumi.core.configs import (
    DataParams,
    DatasetParams,
    DatasetSplitParams,
    ModelParams,
    TrainingConfig,
)


def test_training_config_processor_kwargs():
    """Test that json, regex, and choice parameters are mutually exclusive."""
    config = TrainingConfig(
        model=ModelParams(
            model_name="llava-hf/llava-1.5-7b-hf",
            processor_kwargs={"num_patches": 16},
        ),
        data=DataParams(
            train=DatasetSplitParams(
                collator_name="llava_collator",
                datasets=[
                    DatasetParams(
                        dataset_name="merve/vqav2-small",
                        split="train",
                        dataset_kwargs={"processor_name": "llava-hf/llava-1.5-7b-hf"},
                    ),
                    DatasetParams(
                        dataset_name="HuggingFaceH4/llava-instruct-mix-vsft",
                        split="train",
                        dataset_kwargs={
                            "processor_name": "llava-hf/llava-1.5-7b-hf",
                            "processor_kwargs": {"num_patches": 32, "foo": "bar"},
                        },
                    ),
                    DatasetParams(
                        dataset_name="coco_captions",
                        split="train",
                        dataset_kwargs={
                            "processor_name": "microsoft/Phi-3-vision-128k-instruct"
                        },
                    ),
                    DatasetParams(
                        dataset_name="coco_captions",
                        split="test",
                        dataset_kwargs={
                            "processor_name": "llava-hf/llava-1.5-7b-hf",
                            "processor_kwargs": {},
                        },
                    ),
                ],
            ),
            validation=DatasetSplitParams(
                collator_name="llava_collator",
                datasets=[
                    DatasetParams(
                        dataset_name="merve/vqav2-small",
                        split="validation",
                        dataset_kwargs={"processor_name": "llava-hf/llava-1.5-7b-hf"},
                    )
                ],
            ),
            test=DatasetSplitParams(
                collator_name="llava_collator",
                datasets=[
                    DatasetParams(
                        dataset_name="HuggingFaceH4/llava-instruct-mix-vsft",
                        split="test",
                        dataset_kwargs={"processor_name": "llava-hf/llava-1.5-7b-hf"},
                    )
                ],
            ),
        ),
    )
    assert len(config.data.train.datasets) == 4
    assert config.data.train.datasets[0] == DatasetParams(
        dataset_name="merve/vqav2-small",
        split="train",
        dataset_kwargs={
            "processor_name": "llava-hf/llava-1.5-7b-hf",
            "processor_kwargs": {
                "num_patches": 16,
            },
        },
    )
    assert config.data.train.datasets[1] == DatasetParams(
        dataset_name="HuggingFaceH4/llava-instruct-mix-vsft",
        split="train",
        dataset_kwargs={
            "processor_name": "llava-hf/llava-1.5-7b-hf",
            "processor_kwargs": {"num_patches": 32, "foo": "bar"},
        },
    )
    assert config.data.train.datasets[2] == DatasetParams(
        dataset_name="coco_captions",
        split="train",
        dataset_kwargs={"processor_name": "microsoft/Phi-3-vision-128k-instruct"},
    )
    assert config.data.train.datasets[3] == DatasetParams(
        dataset_name="coco_captions",
        split="test",
        dataset_kwargs={
            "processor_name": "llava-hf/llava-1.5-7b-hf",
            "processor_kwargs": {},
        },
    )

    assert len(config.data.validation.datasets) == 1
    assert config.data.validation.datasets[0] == DatasetParams(
        dataset_name="merve/vqav2-small",
        split="validation",
        dataset_kwargs={
            "processor_name": "llava-hf/llava-1.5-7b-hf",
            "processor_kwargs": {
                "num_patches": 16,
            },
        },
    )

    assert len(config.data.test.datasets) == 1
    assert config.data.test.datasets[0] == DatasetParams(
        dataset_name="HuggingFaceH4/llava-instruct-mix-vsft",
        split="test",
        dataset_kwargs={
            "processor_name": "llava-hf/llava-1.5-7b-hf",
            "processor_kwargs": {
                "num_patches": 16,
            },
        },
    )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic class for using HuggingFace datasets with input/output columns.

Allows users to specify the prompt and response columns at the config level.
"""

from typing import Union

import pandas as pd

from oumi.core.datasets import BaseSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import Conversation, Message, Role


@register_dataset("PromptResponseDataset")
class PromptResponseDataset(BaseSftDataset):
    """Converts HuggingFace Datasets with input/output columns to Message format.

    Example:
        dataset = PromptResponseDataset(hf_dataset_path="O1-OPEN/OpenO1-SFT",
        prompt_column="instruction",
        response_column="output")
    """

    default_dataset = "O1-OPEN/OpenO1-SFT"

    def __init__(
        self,
        *,
        hf_dataset_path: str = "O1-OPEN/OpenO1-SFT",
        prompt_column: str = "instruction",
        response_column: str = "output",
        **kwargs,
    ) -> None:
        """Initializes a new instance of the PromptResponseDataset class."""
        self.prompt_column = prompt_column
        self.response_column = response_column
        kwargs["dataset_name"] = hf_dataset_path
        super().__init__(**kwargs)

    def transform_conversation(self, example: Union[dict, pd.Series]) -> Conversation:
        """Preprocesses the inputs of the example and returns a dictionary.

        Args:
            example (dict or Pandas Series): An example containing `input` (optional),
                `instruction`, and `output` entries.

        Returns:
            dict: The input example converted to messages dictionary format.

        """
        messages = []

        user_prompt = str(example[self.prompt_column])
        messages.append(Message(role=Role.USER, content=user_prompt))

        if self.response_column:
            model_output = str(example[self.response_column])
            messages.append(Message(role=Role.ASSISTANT, content=model_output))

        return Conversation(messages=messages)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import dataclasses
import hashlib
import json
import time
import uuid
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

import jsonlines
from hdrh.histogram import HdrHistogram
from tqdm import tqdm

from oumi.core.configs import (
    GenerationParams,
    InferenceConfig,
    ModelParams,
)
from oumi.core.types.conversation import Conversation
from oumi.utils.logging import logger
from oumi.utils.math_utils import is_power_of_two


class BaseInferenceEngine(ABC):
    """Base class for running model inference."""

    _model_params: ModelParams
    """The model parameters."""

    _generation_params: GenerationParams
    """The generation parameters."""

    def __init__(
        self,
        model_params: ModelParams,
        *,
        generation_params: Optional[GenerationParams] = None,
    ):
        """Initializes the inference engine.

        Args:
            model_params: The model parameters.
            generation_params: The generation parameters.
        """
        self._model_params = copy.deepcopy(model_params)
        if generation_params:
            self._check_unsupported_params(generation_params)
        else:
            generation_params = GenerationParams()
        self._generation_params = generation_params

        self._latency_histogram_online = HdrHistogram(1, 60 * 1000, 1)
        self._latency_histogram_from_file = HdrHistogram(20, 180 * 1000, 1)
        self._dataset_hash = None

    def infer(
        self,
        input: Optional[list[Conversation]] = None,
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference.

        Args:
            input: A list of conversations to run inference on. Optional.
            inference_config: Parameters for inference.
                If not specified, a default config is inferred.

        Returns:
            List[Conversation]: Inference output.
        """
        if input is not None and (
            inference_config and inference_config.input_path is not None
        ):
            raise ValueError(
                "Only one of input or inference_config.input_path should be provided."
            )

        # Ensure the inference config has up-to-date generation parameters.
        if inference_config:
            if inference_config.generation:
                self._check_unsupported_params(inference_config.generation)
            elif self._generation_params:
                inference_config = copy.deepcopy(inference_config)
                inference_config.generation = self._generation_params

                # Warn the user: They provided an inference config without generation
                # params, so what was the point of providing it in the first place?
                logger.warning(
                    "No generation parameters provided in the inference config. Using "
                    "the generation parameters that the engine was initialized with."
                )

        output_path = inference_config.output_path if inference_config else None

        # Load input conversations either from file or use provided input
        conversations_to_process: list[Conversation] = []
        if input is not None:
            conversations_to_process = input
        elif inference_config and inference_config.input_path is not None:
            conversations_to_process = self._read_conversations(
                inference_config.input_path,
            )
        else:
            raise ValueError(
                "One of input or inference_config.input_path must be provided."
            )

        unique_conversation_ids = set()
        for i, conversation in enumerate(conversations_to_process):
            if conversation.conversation_id is not None:
                if conversation.conversation_id in unique_conversation_ids:
                    raise ValueError(
                        f"Conversation ID {conversation.conversation_id} is not unique."
                    )
                unique_conversation_ids.add(conversation.conversation_id)
                continue

            # Generate a deterministic conversation ID based on index if none exists
            messages = conversation.messages
            all_str_message_content = ",".join(
                [
                    message.content
                    for message in messages
                    if isinstance(message.content, str)
                ]
            )
            content_hash = hashlib.sha256(all_str_message_content.encode()).hexdigest()
            id_name = str(i) + "_" + content_hash

            conversation.conversation_id = str(
                uuid.uuid5(
                    uuid.NAMESPACE_DNS,
                    id_name,
                )
            )
            unique_conversation_ids.add(conversation.conversation_id)

        # Calculate the hash of the dataset to use for referencing the inference
        # filename between runs, given the same model and generation parameters.
        dataset_hash = ""
        if len(conversations_to_process) > 0:
            row_hashes = [
                hashlib.sha256(c.to_json().encode()).hexdigest()
                for c in conversations_to_process
            ]
            dataset_hash = hashlib.sha256(",".join(row_hashes).encode()).hexdigest()
        self._dataset_hash = dataset_hash

        completed_conversations = self._load_from_scratch(output_path)

        # Filter out already completed conversations
        remaining_conversations = self._filter_incomplete_conversations(
            conversations_to_process, completed_conversations
        )

        if len(remaining_conversations) < len(conversations_to_process):
            logger.info(
                f"Found {len(completed_conversations)} completed conversations. "
                f"Processing remaining {len(remaining_conversations)} conversations."
            )

        # Run inference only on remaining conversations
        start_time = time.perf_counter()
        histogram = self._latency_histogram_online
        inference_results = self._infer_online(
            remaining_conversations, inference_config
        )
        histogram.record_value((time.perf_counter() - start_time) * 1e3)
        self._maybe_log_latency_histogram(histogram)

        if len(inference_results) == len(conversations_to_process):
            final_results = inference_results
        else:
            # Incomplete inference results were saved to scratch file.
            # Load all results from scratch to get all results.
            final_results = self._load_from_scratch(output_path)
            if len(final_results) != len(conversations_to_process):
                scratch_filepath = self._get_scratch_filepath(output_path)
                raise ValueError(
                    f"Number of final results ({len(final_results)}) does not match "
                    f"number of conversations to process "
                    f"({len(conversations_to_process)}). "
                    f"This typically occurs when the inference configuration has been "
                    f"changed after partial processing. Cached results are stored at: "
                    f"{scratch_filepath}. With the updated caching logic, this should "
                    f"automatically resolve in future runs."
                )

        self._cleanup_scratch_file(output_path)

        sorted_conversations = {
            conv.conversation_id: conv for conv in conversations_to_process
        }

        for conv in final_results:
            sorted_conversations[conv.conversation_id] = conv

        final_results = list(sorted_conversations.values())

        if inference_config and inference_config.output_path:
            self._save_conversations(final_results, inference_config.output_path)

        return final_results

    def _maybe_log_latency_histogram(self, histogram: Optional[HdrHistogram]) -> None:
        """Logs the histogram if it is not None.

        Args:
            histogram: The histogram to log.
        """
        if histogram is None:
            return
        total_count = histogram.get_total_count()
        # TODO: Define a better way to enable/configure this logging.
        if not (
            isinstance(total_count, int)
            and total_count >= 2
            and is_power_of_two(total_count)
        ):
            return

        p50 = histogram.get_value_at_percentile(50)
        p90 = histogram.get_value_at_percentile(90)
        p99 = histogram.get_value_at_percentile(99)
        logger.debug(
            f"{self.__class__.__name__}: "
            f"Latency Histogram: {total_count} samples recorded:"
            f"\tp50: {p50:.1f}ms\tp90: {p90:.1f}ms\tp99: {p99:.1f}ms"
        )

    def _read_conversations(self, input_filepath: str) -> list[Conversation]:
        """Reads conversations from a file in Oumi chat format.

        Args:
            input_filepath: The path to the file containing the conversations.

        Returns:
            List[Conversation]: A list of conversations read from the file.
        """
        conversations = []
        with open(input_filepath) as f:
            for line in f:
                # Only parse non-empty lines.
                if line.strip():
                    conversation = Conversation.from_json(line)
                    conversations.append(conversation)
        return conversations

    def _load_from_scratch(self, output_filepath: Optional[str]) -> list[Conversation]:
        """Loads conversations from a scratch file.

        Args:
            output_filepath: The path to the output file. This is used to determine the
                location of the scratch file.

        Returns:
            list[Conversation]: A list of conversations loaded from the scratch file,
                or an empty list if the scratch file does not exist or is empty.
        """
        scratch_filepath = self._get_scratch_filepath(output_filepath)
        if not Path(scratch_filepath).exists():
            return []

        conversations = []
        with jsonlines.open(scratch_filepath, mode="r") as reader:
            for line in reader:
                conversations.append(Conversation.from_dict(line))
        return conversations

    def _filter_incomplete_conversations(
        self,
        input_conversations: list[Conversation],
        completed_conversations: list[Conversation],
    ) -> list[Conversation]:
        """Filters out conversations that have already been completed.

        Args:
            input_conversations: List of conversations to run inference on
            completed_conversations: List of conversations already completed

        Returns:
            list[Conversation]: List of conversations that still need inference results
        """
        completed_ids = {
            conv.conversation_id
            for conv in completed_conversations
            if conv.conversation_id is not None
        }
        return [
            conv
            for conv in input_conversations
            if conv.conversation_id not in completed_ids
        ]

    def _get_scratch_filepath(self, output_filepath: Optional[str]) -> str:
        """Returns a scratch filepath for the given output filepath.

        The scratch filepath always includes a hash of the model parameters, generation
        parameters, and dataset to ensure cache consistency when configurations change.

        For example, if the output filepath is "/foo/bar/output.json", the scratch
        filepath will be "/foo/bar/scratch/output_<hash>.json"

        If no output filepath is provided, a temporary file is used and placed in
        "~/.cache/oumi/tmp/temp_inference_output_<hash>.jsonl".

        Args:
            output_filepath: The output filepath.

        Returns:
            str: The scratch filepath.
        """
        # Always compute the inference hash to ensure cache consistency
        model_params = self._model_params
        model_params_str = json.dumps(dataclasses.asdict(model_params))
        generation_params = self._generation_params
        generation_params_str = json.dumps(dataclasses.asdict(generation_params))
        inference_hash = hashlib.sha256(
            f"{model_params_str}_{generation_params_str}_{self._dataset_hash}".encode()
        ).hexdigest()

        if output_filepath is not None:
            original_filepath = Path(output_filepath)
            # Include hash in filename to ensure cache consistency
            stem = original_filepath.stem
            suffix = original_filepath.suffix
            hashed_filename = f"{stem}_{inference_hash}{suffix}"
            return str(original_filepath.parent / "scratch" / hashed_filename)

        path_prefix = Path.home() / ".cache" / "oumi" / "tmp"
        return str(path_prefix / f"temp_inference_output_{inference_hash}.jsonl")

    def _save_conversation_to_scratch(
        self, conversation: Conversation, output_filepath: Optional[str]
    ) -> None:
        """Appends a conversation to a file in Oumi chat format.

        Args:
            conversation: The conversation to save.
            output_filepath: The path to the file where the conversation should be
                saved.
        """
        # Make the directory if it doesn't exist.
        scratch_filepath = self._get_scratch_filepath(output_filepath)
        Path(scratch_filepath).parent.mkdir(parents=True, exist_ok=True)
        with jsonlines.open(scratch_filepath, mode="a") as writer:
            json_obj = conversation.to_dict()
            writer.write(json_obj)

    def _cleanup_scratch_file(self, output_filepath: Optional[str]) -> None:
        """Delete the scratch file from the file system if it exists.

        Args:
            output_filepath: The path to the output file. This is used to determine the
                location of the scratch file.
        """
        scratch_filepath = self._get_scratch_filepath(output_filepath)
        if Path(scratch_filepath).exists():
            Path(scratch_filepath).unlink()

    def _save_conversations(
        self, conversations: list[Conversation], output_filepath: str
    ) -> None:
        """Saves conversations to a file in Oumi chat format.

        Args:
            conversations: A list of conversations to save.
            output_filepath: The path to the file where the conversations should be
                saved.
        """
        # Make the directory if it doesn't exist.
        Path(output_filepath).parent.mkdir(parents=True, exist_ok=True)
        with jsonlines.open(output_filepath, mode="w") as writer:
            for conversation in tqdm(conversations, desc="Saving conversations"):
                json_obj = conversation.to_dict()
                writer.write(json_obj)

    def _check_unsupported_params(self, generation_params: GenerationParams):
        """Checks for unsupported parameters and logs warnings.

        If a parameter is not supported, and a non-default value is provided,
        a warning is logged.
        """
        supported_params = self.get_supported_params()
        default_generation_params = GenerationParams()

        for param_name, value in generation_params:
            if param_name not in supported_params:
                is_non_default_value = (
                    getattr(default_generation_params, param_name) != value
                )

                if is_non_default_value:
                    logger.warning(
                        f"{self.__class__.__name__} does not support {param_name}. "
                        f"Received value: {param_name}={value}. "
                        "This parameter will be ignored."
                    )

    @abstractmethod
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine.

        Override this method in derived classes to specify which parameters
        are supported.

        Returns:
            Set[str]: A set of supported parameter names.
        """
        raise NotImplementedError

    @abstractmethod
    def _infer_online(
        self,
        input: list[Conversation],
        inference_config: Optional[InferenceConfig] = None,
    ) -> list[Conversation]:
        """Runs model inference online.

        Args:
            input: A list of conversations to run inference on.
            inference_config: Parameters for inference.

        Returns:
            List[Conversation]: Inference output.
        """
        raise NotImplementedError

    def apply_chat_template(
        self, conversation: Conversation, **tokenizer_kwargs
    ) -> str:
        """Applies the chat template to the conversation.

        Args:
            conversation: The conversation to apply the chat template to.
            tokenizer_kwargs: Additional keyword arguments to pass to the tokenizer.

        Returns:
            str: The conversation with the chat template applied.
        """
        tokenizer = getattr(self, "_tokenizer", None)

        if tokenizer is None:
            raise ValueError("Tokenizer is not initialized.")

        if tokenizer.chat_template is None:
            raise ValueError("Tokenizer does not have a chat template.")

        if "tokenize" not in tokenizer_kwargs:
            tokenizer_kwargs["tokenize"] = False

        return tokenizer.apply_chat_template(conversation, **tokenizer_kwargs)



import os
import sys
from pathlib import Path

from oumi.core.configs import InferenceConfig
from oumi.inference import VLLMInferenceEngine

_INPUT_DIR = "/input_gcs_bucket"
_OUTPUT_DIR = "/output_gcs_bucket"

_INFERENCE_ENV_VAR = "INFERENCE_CONFIG"


def get_new_inputs(source: str, processed: str) -> list[str]:
    """Returns a list of new files in the source directory."""
    source_filenames = {f.name for f in Path(source).rglob("*")}
    processed_filenames = {f.name for f in Path(processed).rglob("*")}
    return list(source_filenames - processed_filenames)


def run_inference(
    file: str,
    input_dir: str,
    output_dir: str,
    config: InferenceConfig,
    engine: VLLMInferenceEngine,
) -> None:
    """Runs inference on a file."""
    config.input_path = str(Path(input_dir) / file)
    config.output_path = str(Path(output_dir) / file)
    print(f"Preparing to read input file: {config.input_path}")
    print(f"Preparing to write output file: {config.output_path}")
    engine.infer(config)


def main():
    """Runs inference on new files in the input directory."""
    parallelism = 1
    if len(sys.argv) > 1:
        parallelism = int(sys.argv[1])
    os.environ["VLLM_WORKER_MULTIPROC_METHOD"] = "spawn"
    print("Checking for new files...")
    new_files = get_new_inputs(_INPUT_DIR, _OUTPUT_DIR)
    yaml_path = os.getenv(_INFERENCE_ENV_VAR)
    if not yaml_path:
        raise ValueError(f"Environment variable {_INFERENCE_ENV_VAR} not set.")
    config = InferenceConfig.from_yaml(yaml_path)
    engine = VLLMInferenceEngine(
        config.model,
        tensor_parallel_size=parallelism,
    )
    for file in new_files:
        print(f"Running inference on {file}...")
        run_inference(file, _INPUT_DIR, _OUTPUT_DIR, config, engine)
        print(f"Inference complete for {file}.")
    print("All files processed!")


if __name__ == "__main__":
    main()



from collections.abc import Sequence

import pytest
from transformers import AutoTokenizer

from oumi.core.registry import REGISTRY
from oumi.core.types.conversation import Conversation, Message


def is_content_empty_expected(dataset_name, conversation_idx, message_idx):
    """Determine if the content of a message is expected to be empty.

    In 99.999% of cases, no message should have empty content. However there are
    some known cases where the content is expected to be empty. This function
    contains a hard-coded list of such known cases.
    """
    known_empty_messages = {
        # This is the answer to the prompt: "write 5 empty sentences"
        ("Magpie-Align/Llama-3-Magpie-Pro-1M-v0.1", 318135, 1),
    }
    return (dataset_name, conversation_idx, message_idx) in known_empty_messages


@pytest.fixture(
    params=[
        "Magpie-Align/Llama-3-Magpie-Pro-1M-v0.1",
        "Magpie-Align/Magpie-Pro-300K-Filtered",
        "argilla/magpie-ultra-v0.1",
        "argilla/databricks-dolly-15k-curated-en",
    ]
)
def dataset_fixture(request):
    dataset_name = request.param
    dataset_class = REGISTRY.get_dataset(dataset_name)
    if dataset_class is None:
        pytest.fail(f"Dataset {dataset_name} not found in registry")
    tokenizer = AutoTokenizer.from_pretrained("gpt2")
    return dataset_name, dataset_class(
        dataset_name=dataset_name, split="train", tokenizer=tokenizer
    )


@pytest.mark.e2e
def test_dataset_conversation(dataset_fixture):
    dataset_name, dataset = dataset_fixture
    assert len(dataset) > 0, f"Dataset {dataset_name} is empty"

    # Iterate through all items in the dataset
    for idx in range(len(dataset)):
        conversation = dataset.conversation(idx)

        # Check the conversation structure
        assert isinstance(conversation, Conversation), (
            f"Conversation at index {idx} is not a Conversation object. "
            f"Type: {type(conversation)}"
        )
        assert len(conversation.messages) > 0, (
            f"Conversation at index {idx} has no messages"
        )

        # Check that each message in the conversation has the expected structure
        for msg_idx, message in enumerate(conversation.messages):
            assert isinstance(message, Message)
            assert message.role in ["user", "assistant"], (
                f"Invalid role for message {msg_idx} in conversation at index {idx}. "
                f"Role: {message.role}"
            )
            assert isinstance(message.content, str)

            if is_content_empty_expected(dataset_name, idx, msg_idx):
                assert message.content == "", (
                    f"Content of message {msg_idx} in conversation at index {idx} "
                    f"is not empty. Content: {message.content}"
                )
            else:
                assert len(message.content) > 0, (
                    f"Content of message {msg_idx} in conversation "
                    f"at index {idx} is empty"
                )
            assert isinstance(message.content, str), (
                f"Type of message {msg_idx} in conversation at index {idx} "
                f"is not 'text'. Type: {type(message.content)}"
            )

        assert conversation.messages[0].role == "user", (
            f"First message in conversation at index {idx} is not from user. "
            f"Role: {conversation.messages[0].role}"
        )
        assert conversation.messages[-1].role == "assistant", (
            f"Last message in conversation at index {idx} is not from assistant. "
            f"Role: {conversation.messages[-1].role}"
        )


@pytest.mark.skip(
    reason="This test is very time consuming, and should be run manually."
)
def test_dataset_prompt_generation(dataset_fixture):
    dataset_name, dataset = dataset_fixture
    assert len(dataset) > 0, f"Dataset {dataset_name} is empty"

    for idx in range(len(dataset)):
        prompt = dataset.prompt(idx)
        assert isinstance(prompt, str), (
            f"Prompt at index {idx} is not a string. Type: {type(prompt)}"
        )
        assert len(prompt) > 0, f"Prompt at index {idx} is empty"


@pytest.mark.skip(
    reason="This test is very time consuming, and should be run manually."
)
def test_dataset_model_inputs(dataset_fixture):
    dataset_name, dataset = dataset_fixture
    assert len(dataset) > 0, f"Dataset {dataset_name} is empty"

    # Iterate through all items in the dataset
    for idx in range(len(dataset)):
        item = dataset[idx]

        # Check that each item has the expected keys
        assert "input_ids" in item, f"'input_ids' not found in item at index {idx}"
        assert isinstance(item["input_ids"], Sequence), (
            f"'input_ids' is not a Sequence at index {idx}. "
            f"Type: {type(item['input_ids'])}"
        )



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from oumi.core.configs import JobConfig
from oumi.core.launcher import BaseCloud, BaseCluster, JobStatus
from oumi.core.registry import REGISTRY, RegistryType


class Launcher:
    """A class for managing the lifecycle of jobs on different clouds."""

    def __init__(self):
        """Initializes a new instance of the Launcher class."""
        self._clouds: dict[str, BaseCloud] = dict()

    def _initialize_new_clouds(self) -> None:
        """Initializes new clouds. Existing clouds are not re-initialized."""
        for name, builder in REGISTRY.get_all(RegistryType.CLOUD).items():
            if name not in self._clouds:
                self._clouds[name] = builder()

    def _get_cloud_by_name(self, cloud: str) -> BaseCloud:
        """Gets the cloud instance for the specified cloud name."""
        if cloud not in self._clouds:
            cloud_builder = REGISTRY.get(cloud, RegistryType.CLOUD)
            if not cloud_builder:
                raise ValueError(f"Cloud {cloud} not found in the registry.")
            self._clouds[cloud] = cloud_builder()
        return self._clouds[cloud]

    def cancel(self, job_id: str, cloud_name: str, cluster_name: str) -> JobStatus:
        """Cancels the specified job."""
        cloud = self._get_cloud_by_name(cloud_name)
        cluster = cloud.get_cluster(cluster_name)
        if not cluster:
            raise ValueError(f"Cluster {cluster_name} not found.")
        return cluster.cancel_job(job_id)

    def down(self, cloud_name: str, cluster_name: str) -> None:
        """Turns down the specified cluster."""
        cloud = self._get_cloud_by_name(cloud_name)
        cluster = cloud.get_cluster(cluster_name)
        if not cluster:
            raise ValueError(f"Cluster {cluster_name} not found.")
        cluster.down()

    def get_cloud(self, job_or_cloud: Union[JobConfig, str]) -> BaseCloud:
        """Gets the cloud instance for the specified job."""
        if isinstance(job_or_cloud, str):
            return self._get_cloud_by_name(job_or_cloud)
        return self._get_cloud_by_name(job_or_cloud.resources.cloud)

    def run(self, job: JobConfig, cluster_name: str) -> JobStatus:
        """Runs the specified job on the specified cluster.

        Args:
            job: The job configuration.
            cluster_name: The name of the cluster to run the job on.

        Returns:
            Optional[JobStatus]: The status of the job.
        """
        cloud = self.get_cloud(job)
        cluster = cloud.get_cluster(cluster_name)
        if not cluster:
            raise ValueError(f"Cluster {cluster_name} not found.")
        return cluster.run_job(job)

    def status(
        self,
        cloud: Optional[str] = None,
        cluster: Optional[str] = None,
        id: Optional[str] = None,
    ) -> dict[str, list[JobStatus]]:
        """Gets the status of all jobs across all clusters.

        Args:
            cloud: If specified, filters all jobs to only those on the specified cloud.
            cluster: If specified, filters all jobs to only those on the specified
                cluster.
            id: If specified, filters all jobs to only those with the specified ID.

        Returns:
            dict[str, list(JobStatus)]: The status of all jobs, indexed by cloud name.
        """
        # Pick up any newly registered cloud builders.
        self._initialize_new_clouds()
        statuses: dict[str, list[JobStatus]] = {}
        for cloud_name, target_cloud in self._clouds.items():
            # Ignore clouds not matching the filter criteria.
            if cloud and cloud_name != cloud:
                continue
            statuses[cloud_name] = []
            for target_cluster in target_cloud.list_clusters():
                # Ignore clusters not matching the filter criteria.
                if cluster and target_cluster.name() != cluster:
                    continue
                for job in target_cluster.get_jobs():
                    # Ignore jobs not matching the filter criteria.
                    if id and job.id != id:
                        continue
                    statuses[cloud_name].append(job)
        return statuses

    def stop(self, cloud_name: str, cluster_name: str) -> None:
        """Stops the specified cluster."""
        cloud = self._get_cloud_by_name(cloud_name)
        cluster = cloud.get_cluster(cluster_name)
        if not cluster:
            raise ValueError(f"Cluster {cluster_name} not found.")
        cluster.stop()

    def up(
        self, job: JobConfig, cluster_name: Optional[str], **kwargs
    ) -> tuple[BaseCluster, JobStatus]:
        """Creates a new cluster and starts the specified job on it."""
        cloud = self.get_cloud(job)
        job_status = cloud.up_cluster(job, cluster_name, **kwargs)
        cluster = cloud.get_cluster(job_status.cluster)
        if not cluster:
            raise RuntimeError(f"Cluster {job_status.cluster} not found.")
        return (cluster, job_status)

    def which_clouds(self) -> list[str]:
        """Gets the names of all clouds in the registry."""
        return [name for name, _ in REGISTRY.get_all(RegistryType.CLOUD).items()]


LAUNCHER = Launcher()
# Explicitly expose the public methods of the default Launcher instance.
#: A convenience method for Launcher.cancel.
cancel = LAUNCHER.cancel
#: A convenience method for Launcher.down.
down = LAUNCHER.down
#: A convenience method for Launcher.get_cloud.
get_cloud = LAUNCHER.get_cloud
#: A convenience method for Launcher.run.
run = LAUNCHER.run
#: A convenience method for Launcher.status.
status = LAUNCHER.status
#: A convenience method for Launcher.stop.
stop = LAUNCHER.stop
#: A convenience method for Launcher.up.
up = LAUNCHER.up
#: A convenience method for Launcher.which_clouds.
which_clouds = LAUNCHER.which_clouds



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
from typing import Any, Optional

from tqdm.asyncio import tqdm
from typing_extensions import override

from oumi.core.configs import GenerationParams, ModelParams, RemoteParams
from oumi.core.types.conversation import Conversation, Message, Role
from oumi.inference.adaptive_semaphore import PoliteAdaptiveSemaphore
from oumi.inference.remote_inference_engine import RemoteInferenceEngine
from oumi.utils.logging import logger

try:
    import boto3  # pyright: ignore[reportMissingImports]
    from botocore.exceptions import ClientError  # pyright: ignore[reportMissingImports]
except ModuleNotFoundError:
    boto3 = None  # type: ignore
    ClientError = None  # type: ignore

_CONTENT_KEY: str = "content"
_ROLE_KEY: str = "role"
_AWS_REGION_ENV_VAR: str = "AWS_REGION"


class BedrockInferenceEngine(RemoteInferenceEngine):
    """Inference engine for running inference against the Bedrock API.

    This class extends RemoteInferenceEngine to provide specific functionality
    for interacting with Bedrock's language models via their API. It handles
    the conversion of Oumi's Conversation objects to Bedrock's expected input
    format, as well as parsing the API responses back into Conversation objects.

    Note:
        This engine requires the boto3 package to be installed.
        If not installed, it will raise a RuntimeError.
    """

    def __init__(
        self,
        model_params: ModelParams,
        *,
        generation_params: Optional[GenerationParams] = None,
        remote_params: Optional[RemoteParams] = None,
    ):
        """Initializes the BedrockInferenceEngine.

        Args:
            model_params: Parameters for the model.
            generation_params: Parameters for generation.
            remote_params: Parameters for remote inference.

        Raises:
            RuntimeError: If the boto3 package is not installed.
        """
        if not boto3:
            raise RuntimeError(
                "boto3 is not installed. Please install it with 'pip install boto3'."
            )

        super().__init__(
            model_params=model_params,
            generation_params=generation_params,
            remote_params=remote_params,
        )

    @property
    @override
    def base_url(self) -> Optional[str]:
        """Return the default base URL for the Bedrock API."""
        return None

    @property
    @override
    def api_key_env_varname(self) -> Optional[str]:
        """Return the default environment variable name for the Bedrock API key."""
        return None

    def _bedrock_client(self, remote_params: RemoteParams) -> Any:
        region = os.getenv(_AWS_REGION_ENV_VAR)
        if not region:
            raise ValueError(f"Environment variable {_AWS_REGION_ENV_VAR} not set.")
        return boto3.client("bedrock-runtime", region_name=region)  # type: ignore

    @override
    def _convert_conversation_to_api_input(
        self,
        conversation: Conversation,
        generation_params: GenerationParams,
        model_params: ModelParams,
    ) -> dict[str, Any]:
        """Converts a conversation to a Bedrock API input.

        Args:
            conversation: The conversation to convert.
            generation_params: Parameters for text generation.
            model_params: Model parameters to use during inference.

        Returns:
            Dict[str, Any]: A dictionary containing the formatted input for the
                Bedrock API, including the model, messages, and generation parameters.
        """
        system_messages = [
            message for message in conversation.messages if message.role == Role.SYSTEM
        ]
        if len(system_messages) > 0:
            system_message = system_messages[0].content

            if len(system_messages) > 1:
                logger.warning(
                    "Multiple system messages found in conversation. "
                    "Only using the first one."
                )
        else:
            system_message = None

        messages = [
            message for message in conversation.messages if message.role != Role.SYSTEM
        ]

        body = {
            "inferenceConfig": {
                "maxTokens": generation_params.max_new_tokens,
                "temperature": generation_params.temperature,
                "topP": generation_params.top_p,
            },
            "messages": self._to_bedrock_messages(messages),
        }

        if system_message:
            body["system"] = [{"text": system_message}]

        if model_params.model_kwargs:
            body["additionalModelRequestFields"] = model_params.model_kwargs

        if generation_params.stop_strings:
            body["inferenceConfig"]["stopSequences"] = generation_params.stop_strings

        return body

    def _to_bedrock_messages(self, messages: list[Message]) -> list[dict]:
        result = []
        for m in messages:
            # Map Oumi roles to Bedrock roles
            if m.role == Role.USER:
                bedrock_role = "user"
            elif m.role == Role.ASSISTANT:
                bedrock_role = "assistant"
            else:
                logger.warning(f"Skipping message with unsupported role: {m.role}")
                continue

            content_blocks: list[dict] = []
            if m.contains_single_text_content_item_only():
                content_blocks = [{"text": m.text_content_items[0].content}]
            elif m.contains_images():
                img = m.image_content_items[0]
                if getattr(img, "binary", None):
                    content_blocks = [{"image": {"source": {"bytes": img.binary}}}]
                else:
                    uri = img.content or ""
                    if uri.startswith("s3://"):
                        content_blocks = [
                            {"image": {"source": {"s3Location": {"uri": uri}}}}
                        ]
                    else:
                        content_blocks = [{"image": {"source": {"url": uri}}}]
            else:
                content_blocks = [{"text": m.compute_flattened_text_content()}]

            result.append({"role": bedrock_role, "content": content_blocks})
        return result

    @override
    def _default_remote_params(self) -> RemoteParams:
        """Returns the default remote parameters."""
        return RemoteParams()

    @override
    def _set_required_fields_for_inference(self, remote_params: RemoteParams):
        """Override to skip API key validation since Bedrock uses AWS credentials."""
        pass

    async def _infer(
        self,
        input: list[Conversation],
        inference_config: Optional[Any] = None,
    ) -> list[Conversation]:
        """Async inference implementation that doesn't use HTTP sessions."""
        semaphore = PoliteAdaptiveSemaphore(
            capacity=self._remote_params.num_workers,
            politeness_policy=self._remote_params.politeness_policy,
        )

        # Create tasks for all conversations
        tasks = [
            self._query_api(
                conversation,
                semaphore,
                None,  # No HTTP session needed for boto3
                inference_config=inference_config,
            )
            for conversation in input
        ]

        disable_tqdm = len(tasks) < 2
        results = await tqdm.gather(*tasks, disable=disable_tqdm)
        return results

    def _call_bedrock_converse(
        self,
        remote_params: RemoteParams,
        model_params: ModelParams,
        body: dict[str, Any],
    ) -> dict[str, Any]:
        """Synchronously invokes Bedrock Converse via boto3."""
        client = self._bedrock_client(remote_params)
        kwargs: dict[str, Any] = {
            "modelId": model_params.model_name,
            "messages": body["messages"],
        }
        if "system" in body:
            kwargs["system"] = body["system"]
        if "inferenceConfig" in body:
            kwargs["inferenceConfig"] = body["inferenceConfig"]
        if "additionalModelRequestFields" in body:
            kwargs["additionalModelRequestFields"] = body[
                "additionalModelRequestFields"
            ]
        return client.converse(**kwargs)

    @override
    async def _query_api(
        self,
        conversation: Conversation,
        semaphore: PoliteAdaptiveSemaphore,
        session: Any,
        inference_config: Optional[Any] = None,
    ) -> Conversation:
        """Queries Bedrock Converse using boto3 instead of HTTP."""
        if inference_config is None:
            remote_params = self._remote_params
            generation_params = self._generation_params
            model_params = self._model_params
            output_path = None
        else:
            remote_params = inference_config.remote_params or self._remote_params
            generation_params = inference_config.generation or self._generation_params
            model_params = inference_config.model or self._model_params
            output_path = inference_config.output_path

        semaphore_or_controller = (
            self._adaptive_concurrency_controller
            if self._remote_params.use_adaptive_concurrency
            else semaphore
        )
        async with semaphore_or_controller:
            api_input = self._convert_conversation_to_api_input(
                conversation, generation_params, model_params
            )
            failure_reason: Optional[str] = None
            for attempt in range(remote_params.max_retries + 1):
                try:
                    if attempt > 0:
                        delay = min(
                            remote_params.retry_backoff_base * (2 ** (attempt - 1)),
                            remote_params.retry_backoff_max,
                        )
                        await asyncio.sleep(delay)

                    response = await asyncio.to_thread(
                        self._call_bedrock_converse,
                        remote_params,
                        model_params,
                        api_input,
                    )
                    result = self._convert_api_output_to_conversation(
                        response, conversation
                    )
                    if output_path:
                        self._save_conversation_to_scratch(result, output_path)
                    await self._try_record_success()
                    return result
                except ClientError as e:  # type: ignore
                    # Capture AWS error message for logging/propagation
                    failure_reason = e.response.get("Error", {}).get("Message") or str(
                        e
                    )
                    await self._try_record_error()
                    if attempt >= remote_params.max_retries:
                        raise RuntimeError(failure_reason) from e
                    continue
                except RuntimeError:
                    raise
                except Exception as e:
                    failure_reason = f"Unexpected error: {str(e)}"
                    await self._try_record_error()
                    if attempt >= remote_params.max_retries:
                        raise RuntimeError(failure_reason) from e
                    continue

            raise RuntimeError(
                f"Failed to query Bedrock after {remote_params.max_retries} retries. "
                + (f"Reason: {failure_reason}" if failure_reason else "")
            )

    @override
    def _convert_api_output_to_conversation(
        self, response: dict[str, Any], original: Conversation
    ) -> Conversation:
        text = ""
        msg = response.get("output", {}).get("message", {})
        for block in msg.get("content", []):
            if "text" in block:
                text += block["text"]
        new_message = Message(content=text, role=Role.ASSISTANT)
        return Conversation(
            messages=[*original.messages, new_message],
            metadata=original.metadata,
            conversation_id=original.conversation_id,
        )

    @override
    def get_supported_params(self) -> set[str]:
        """Returns a set of supported generation parameters for this engine."""
        return {
            "max_new_tokens",
            "stop_strings",
            "temperature",
            "top_p",
        }

    # Override batch inference methods to indicate they're not supported
    @override
    def infer_batch(
        self,
        conversations: list[Conversation],
        inference_config: Optional[Any] = None,
    ) -> str:
        """Bedrock does not support batch inference via OpenAI-style batch API."""
        raise NotImplementedError("Batch inference is not supported for Bedrock API.")

    @override
    def get_batch_status(self, batch_id: str) -> Any:
        """Bedrock does not support batch inference via OpenAI-style batch API."""
        raise NotImplementedError("Batch inference is not supported for Bedrock API.")

    @override
    def list_batches(
        self,
        after: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> Any:
        """Bedrock does not support batch inference via OpenAI-style batch API."""
        raise NotImplementedError("Batch inference is not supported for Bedrock API.")

    @override
    def get_batch_results(
        self,
        batch_id: str,
        conversations: list[Conversation],
    ) -> list[Conversation]:
        """Bedrock does not support batch inference via OpenAI-style batch API."""
        raise NotImplementedError("Batch inference is not supported for Bedrock API.")



import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
import typer
from typer.testing import CliRunner

from oumi.cli.fetch import fetch

runner = CliRunner()


@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command()(fetch)
    return fake_app


@pytest.fixture
def mock_fetch():
    with patch("oumi.cli.fetch.resolve_and_fetch_config") as fetch_mock:
        yield fetch_mock


def test_fetch_with_oumi_prefix_and_explicit_output_dir(app, mock_fetch):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"
        mock_fetch.return_value = expected_path
        # When
        result = runner.invoke(app, [config_path, "-o", str(output_dir)])

        # Then
        assert result.exit_code == 0
        mock_fetch.assert_called_once_with(config_path, output_dir, False)


def test_fetch_without_prefix_and_explicit_output_dir(app, mock_fetch):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"
        mock_fetch.return_value = expected_path
        # When
        result = runner.invoke(app, [config_path, "-o", str(output_dir)])

        # Then
        assert result.exit_code == 0
        mock_fetch.assert_called_once_with("oumi://" + config_path, output_dir, False)


def test_fetch_without_prefix_and_explicit_output_dir_force(app, mock_fetch):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"
        mock_fetch.return_value = expected_path
        # When
        result = runner.invoke(app, [config_path, "-o", str(output_dir), "-f"])

        # Then
        assert result.exit_code == 0
        mock_fetch.assert_called_once_with("oumi://" + config_path, output_dir, True)


def test_fetch_with_oumi_prefix_no_output(app, mock_fetch):
    # Given
    config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
    expected_path = Path(config_path)
    mock_fetch.return_value = expected_path
    # When
    result = runner.invoke(app, [config_path])

    # Then
    assert result.exit_code == 0
    mock_fetch.assert_called_once_with(config_path, None, False)


def test_fetch_without_oumi_prefix_no_output(app, mock_fetch):
    # Given
    config_path = "configs/recipes/smollm/inference/135m_infer.yaml"
    expected_path = Path(config_path)
    mock_fetch.return_value = expected_path
    # When
    result = runner.invoke(app, [config_path])

    # Then
    assert result.exit_code == 0
    mock_fetch.assert_called_once_with("oumi://" + config_path, None, False)


def test_fetch_without_oumi_prefix_no_output_force(app, mock_fetch):
    # Given
    config_path = "configs/recipes/smollm/inference/135m_infer.yaml"
    expected_path = Path(config_path)
    mock_fetch.return_value = expected_path
    # When
    result = runner.invoke(app, [config_path, "-f"])

    # Then
    assert result.exit_code == 0
    mock_fetch.assert_called_once_with("oumi://" + config_path, None, True)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from pathlib import Path
from subprocess import PIPE, Popen
from threading import Lock, Thread
from typing import Optional

from oumi.core.configs import JobConfig
from oumi.core.launcher import JobState, JobStatus


@dataclass
class _LocalJob:
    """A class representing a job running locally."""

    status: JobStatus
    config: JobConfig
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class _JobState(Enum):
    """An enumeration of the possible states of a job."""

    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELED = "CANCELED"


class LocalClient:
    """A client for running jobs locally in a subprocess."""

    # The maximum number of characters to read from the subprocess's stdout and stderr.
    _MAX_BUFFER_SIZE = 1024
    # The environment variable used to specify the logging directory.
    _OUMI_LOGGING_DIR = "OUMI_LOGGING_DIR"

    def __init__(self):
        """Initializes a new instance of the LocalClient class."""
        self._mutex = Lock()
        self._next_job_id = 0
        # A mapping of job IDs to their respective job configurations.
        self._jobs = {}
        self._running_process = None
        self._worker = Thread(target=self._worker_loop, daemon=True)
        self._worker.start()

    def _get_job_state(self, job_state: _JobState) -> JobState:
        """Gets the state of the job."""
        if job_state == _JobState.QUEUED:
            return JobState.PENDING
        elif job_state == _JobState.RUNNING:
            return JobState.RUNNING
        elif job_state == _JobState.COMPLETED:
            return JobState.SUCCEEDED
        elif job_state == _JobState.FAILED:
            return JobState.FAILED
        elif job_state == _JobState.CANCELED:
            return JobState.CANCELLED
        raise ValueError(f"Invalid job state: {job_state}")

    def _update_job_status(self, job_id: str, status: _JobState) -> None:
        """Updates the status of the job. Assumes the mutex is already acquired."""
        if job_id not in self._jobs:
            return
        self._jobs[job_id].status.status = status.value
        self._jobs[job_id].status.state = self._get_job_state(status)
        is_done = status in (_JobState.COMPLETED, _JobState.FAILED, _JobState.CANCELED)
        self._jobs[job_id].status.done = is_done

    def _worker_run_job(self) -> Optional[_LocalJob]:
        """Kicks off and returns a new job. Assumes the mutex is already acquired."""
        job = self._get_next_job()
        if job is None:
            return None
        env_copy = os.environ.copy()
        env_copy.update(job.config.envs)
        # Check if the user has specified a logging directory.
        if self._OUMI_LOGGING_DIR in env_copy:
            logging_dir = Path(env_copy[self._OUMI_LOGGING_DIR])
            logging_dir.mkdir(parents=True, exist_ok=True)
            dt = datetime.now()
            log_format = f"{dt:%Y_%m_%d_%H_%M_%S}_{dt.microsecond // 1000:03d}"
            job.stderr = str(logging_dir / f"{log_format}_{job.status.id}.stderr")
            job.stdout = str(logging_dir / f"{log_format}_{job.status.id}.stdout")
        # Always change to the working directory before running the job.
        working_dir_cmd = f"cd {job.config.working_dir}"
        setup_cmds = job.config.setup or ""
        cmds = "\n".join([working_dir_cmd, setup_cmds, job.config.run])
        # Start the job but don't block.
        stderr_logs = open(job.stderr, "w") if job.stderr else PIPE
        stdout_logs = open(job.stdout, "w") if job.stdout else PIPE
        self._running_process = Popen(
            cmds,
            shell=True,
            env=env_copy,
            stdout=stdout_logs,
            stderr=stderr_logs,
        )
        self._update_job_status(job.status.id, _JobState.RUNNING)
        return job

    def _worker_handle_running_job(self, job: _LocalJob) -> None:
        """Polls and handles the specified job. Acquires the mutex."""
        # Return immediately if no job is running.
        if self._running_process is None:
            return
        # Wait for the job to finish. No need to grab the mutex here.
        if self._running_process.wait() == 0:
            # Job was successful.
            finish_time = datetime.fromtimestamp(time.time()).isoformat()
            with self._mutex:
                self._jobs[
                    job.status.id
                ].status.metadata = f"Job finished at {finish_time} ."
                self._update_job_status(job.status.id, _JobState.COMPLETED)
                if job.stdout is not None:
                    self._jobs[
                        job.status.id
                    ].status.metadata += f" Logs available at: {job.stdout}"
        else:
            # Job failed.
            with self._mutex:
                self._update_job_status(job.status.id, _JobState.FAILED)
                if job.stderr is not None:
                    self._jobs[
                        job.status.id
                    ].status.metadata = f"Error logs available at: {job.stderr}"
                else:
                    error_metadata = ""
                    if self._running_process.stderr is not None:
                        for line in self._running_process.stderr:
                            error_metadata += str(line)
                    # Only keep the last _MAX_BUFFER_SIZE characters.
                    error_metadata = error_metadata[-self._MAX_BUFFER_SIZE :]
                    self._jobs[job.status.id].status.metadata = error_metadata

    def _worker_loop(self):
        """The main worker loop that runs jobs."""
        while True:
            with self._mutex:
                # Run the next job if it exists.
                job = self._worker_run_job()
            # No job to run, sleep for a bit.
            if job is None:
                time.sleep(5)
                continue
            # Wait for the job to finish.
            self._worker_handle_running_job(job)
            # Clear the running process.
            with self._mutex:
                self._running_process = None

    def _generate_next_job_id(self) -> str:
        """Gets the next job ID."""
        job_id = self._next_job_id
        self._next_job_id += 1
        return str(job_id)

    def _get_next_job(self) -> Optional[_LocalJob]:
        """Gets the next QUEUED job from the queue."""
        queued_jobs = [
            job
            for job in self._jobs.values()
            if job.status.status == _JobState.QUEUED.value
        ]
        if len(queued_jobs) == 0:
            return None
        next_job_id = queued_jobs[0].status.id
        for job in queued_jobs:
            if int(job.status.id) < int(next_job_id):
                next_job_id = job.status.id
        return self._jobs[next_job_id]

    def submit_job(self, job: JobConfig) -> JobStatus:
        """Runs the specified job on this cluster."""
        with self._mutex:
            job_id = self._generate_next_job_id()
            name = job.name if job.name else job_id
            status = JobStatus(
                name=name,
                id=job_id,
                status=_JobState.QUEUED.value,
                cluster="",
                metadata="",
                done=False,
                state=JobState.PENDING,
            )
            self._jobs[job_id] = _LocalJob(status=status, config=job)
            return status

    def list_jobs(self) -> list[JobStatus]:
        """Returns a list of job statuses."""
        with self._mutex:
            return [job.status for job in self._jobs.values()]

    def get_job(self, job_id: str) -> Optional[JobStatus]:
        """Gets the specified job's status.

        Args:
            job_id: The ID of the job to get.

        Returns:
            The job status if found, None otherwise.
        """
        job_list = self.list_jobs()
        for job in job_list:
            if job.id == job_id:
                return job
        return None

    def cancel(self, job_id) -> Optional[JobStatus]:
        """Cancels the specified job.

        Args:
            job_id: The ID of the job to cancel.
            queue: The name of the queue to search.

        Returns:
            The job status if found, None otherwise.
        """
        with self._mutex:
            if job_id not in self._jobs:
                return None
            job = self._jobs[job_id]
            if job.status.status == _JobState.RUNNING.value:
                if self._running_process is not None:
                    self._running_process.terminate()
                self._update_job_status(job_id, _JobState.CANCELED)
            elif job.status.status == _JobState.QUEUED.value:
                self._update_job_status(job_id, _JobState.CANCELED)
            return job.status



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, cast

import numpy as np
from typing_extensions import override

from oumi.core.datasets.base_map_dataset import BaseMapDataset
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types.conversation import Role

_PROMPT_KEY = "prompt"
_CHOSEN_KEY = "chosen"
_REJECTED_KEY = "rejected"

_ROLE = "role"
_CONTENT = "content"
_ASSISTANT = "assistant"


class BaseDpoDataset(BaseMapDataset):
    """Preprocess the samples to the Oumi format."""

    def __init__(
        self,
        *,
        dataset_name: Optional[str] = None,
        dataset_path: Optional[str] = None,
        split: Optional[str] = None,
        tokenizer: Optional[BaseTokenizer] = None,
        return_tensors: bool = False,
        **kwargs,
    ) -> None:
        """Initializes a new instance of the BaseDpoDataset class.

        The dataset expects data in the format::

            {
                "prompt": "How is the weather in Tokyo?",
                "chosen": [{"role": "assistant", "content": "It's sunny and warm."}],
                "rejected": [{"role": "assistant", "content": "It's rainy and cold."}]
            }

            OR

            {
                "prompt": "How is the weather in Tokyo?",
                "chosen": "preferred response",
                "rejected": "rejected response"
            }
        """
        super().__init__(
            dataset_name=dataset_name,
            dataset_path=dataset_path,
            split=split,
            **kwargs,
        )

        self._tokenizer = tokenizer
        self._return_tensors = return_tensors

        self._data = self._load_data()

    def transform_preference(self, samples: dict) -> dict:
        """Transform the samples to the Oumi format."""
        prompt = samples[_PROMPT_KEY]
        chosen_chat = samples[_CHOSEN_KEY]
        rejected_chat = samples[_REJECTED_KEY]

        return {
            _PROMPT_KEY: self._to_messages_list(prompt, Role.USER),
            _CHOSEN_KEY: self._to_messages_list(chosen_chat, Role.ASSISTANT),
            _REJECTED_KEY: self._to_messages_list(rejected_chat, Role.ASSISTANT),
        }

    def _process_sample(
        self,
        features,
    ):
        """Tokenize a row of the dataset.

        Example:
        ```python
        >>> from transformers import GPT2Tokenizer
        >>> tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
        >>> features = {"prompt": "The sky is", "chosen": " blue", "rejected": " green"}
        >>> DPOTrainer.tokenize_row(
        ...     features, tokenizer, max_prompt_length=3, max_completion_length=3,
        ...     add_special_tokens=False,
        ... )
        {'prompt_input_ids': [464, 6766, 318], 'chosen_input_ids': [4171, 50256],
        'rejected_input_ids': [4077, 50256]}
        ```
        """
        if self._tokenizer is None:
            raise ValueError("Tokenizer is required to process a sample.")

        # Apply the chat template to the prompt.
        prompt = self._tokenizer.apply_chat_template(features["prompt"], tokenize=False)
        prompt = cast(str, prompt)

        # Apply the chat template to the chosen and rejected turns.
        # To get only the completion part, we tokenizer the prompt + chosen/rejected
        # and then remove the prompt prefix.
        prompt_chosen = self._tokenizer.apply_chat_template(
            features["prompt"] + features["chosen"], tokenize=False
        )
        prompt_chosen = cast(str, prompt_chosen)
        chosen = prompt_chosen[len(prompt) :]

        prompt_rejected = self._tokenizer.apply_chat_template(
            features["prompt"] + features["rejected"], tokenize=False
        )
        prompt_rejected = cast(str, prompt_rejected)
        rejected = prompt_rejected[len(prompt) :]

        # Tokenize the prompt, chosen, and rejected turns.
        prompt_input_ids = self._tokenizer(prompt, add_special_tokens=False)[
            "input_ids"
        ]
        chosen_input_ids = self._tokenizer(chosen, add_special_tokens=False)[
            "input_ids"
        ]
        chosen_input_ids = cast(list[int], chosen_input_ids)
        rejected_input_ids = self._tokenizer(rejected, add_special_tokens=False)[
            "input_ids"
        ]
        rejected_input_ids = cast(list[int], rejected_input_ids)
        chosen_input_ids = chosen_input_ids + [self._tokenizer.eos_token_id]
        rejected_input_ids = rejected_input_ids + [self._tokenizer.eos_token_id]

        return {
            "prompt_input_ids": prompt_input_ids,
            "chosen_input_ids": chosen_input_ids,
            "rejected_input_ids": rejected_input_ids,
        }

    @override
    def transform(self, sample: dict) -> dict:
        """Transform the samples to the Oumi format."""
        return self._process_sample(self.transform_preference(sample))

    def _to_messages_list(
        self, turn: Union[str, dict, list[dict]], role: Role
    ) -> list[dict]:
        """Convert a turn to a conversation dictionary."""
        if isinstance(turn, str):
            return [{"role": role.value, "content": turn}]
        if isinstance(turn, dict):
            return [turn]
        if isinstance(turn, np.ndarray):
            return list(turn)
        if isinstance(turn, list):
            return turn

        raise ValueError(f"Invalid turn type: {type(turn)}")


class BaseExperimentalDpoDataset(BaseDpoDataset):
    """Preprocess the samples to the Oumi format.

    Warning:
        This class is experimental and subject to change.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initializes a new instance of the BaseExperimentalDpoDataset class."""
        from oumi.utils.logging import logger

        logger.warning(
            "`BaseExperimentalDpoDataset` is deprecated and will be removed in the "
            "future. Please use `BaseDpoDataset` instead."
        )

        super().__init__(*args, **kwargs)



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oumi.core.datasets import BasePretrainingDataset
from oumi.core.registry import register_dataset


@register_dataset("PleIAs/YouTube-Commons")
class YouTubeCommonsDataset(BasePretrainingDataset):
    """YouTube-Commons Dataset.

    This dataset is a collection of audio transcripts from 2,063,066 videos shared on
    YouTube under a CC-By license. It contains 22,709,724 original and automatically
    translated transcripts from 3,156,703 videos (721,136 individual channels),
    representing nearly 45 billion words.

    The corpus is multilingual, with a majority of English-speaking content (71%) for
    original languages. Automated translations are provided for nearly all videos in
    English, French, Spanish, German, Russian, Italian, and Dutch.

    This dataset aims to expand the availability of conversational data for research
    in AI, computational social science, and digital humanities.

    See Also:
        - Hugging Face Hub: https://huggingface.co/datasets/PleIAs/YouTube-Commons

    Data Fields:
        - video_id: string
        - video_link: string
        - title: string
        - text: string
        - channel: string
        - channel_id: string
        - date: string
        - license: string
        - original_language: string
        - source_language: string
        - transcription_language: string
        - word_count: int64
        - character_count: int64

    Note:
        The text can be used for training models and republished for reproducibility
        purposes. In accordance with the CC-By license, every YouTube channel is fully
        credited.

    Note:
        This dataset is licensed under CC-BY-4.0.
    """

    default_dataset = "PleIAs/YouTube-Commons"



import os

import pytest

from oumi.utils.distributed_utils import is_using_accelerate, is_using_accelerate_fsdp


def test_is_using_accelerate():
    for var in [
        "ACCELERATE_DYNAMO_BACKEND",
        "ACCELERATE_DYNAMO_MODE",
        "ACCELERATE_DYNAMO_USE_FULLGRAPH",
        "ACCELERATE_DYNAMO_USE_DYNAMIC",
    ]:
        if var in os.environ:
            del os.environ[var]
    assert not is_using_accelerate()

    os.environ["ACCELERATE_DYNAMO_BACKEND"] = "some_value"
    assert is_using_accelerate()
    del os.environ["ACCELERATE_DYNAMO_BACKEND"]

    os.environ["ACCELERATE_DYNAMO_MODE"] = "some_value"
    assert is_using_accelerate()
    del os.environ["ACCELERATE_DYNAMO_MODE"]

    os.environ["ACCELERATE_DYNAMO_USE_FULLGRAPH"] = "some_value"
    assert is_using_accelerate()
    del os.environ["ACCELERATE_DYNAMO_USE_FULLGRAPH"]

    os.environ["ACCELERATE_DYNAMO_USE_DYNAMIC"] = "some_value"
    assert is_using_accelerate()
    del os.environ["ACCELERATE_DYNAMO_USE_DYNAMIC"]

    os.environ["ACCELERATE_DYNAMO_BACKEND"] = "some_value"
    os.environ["ACCELERATE_DYNAMO_MODE"] = "some_value"
    os.environ["ACCELERATE_DYNAMO_USE_FULLGRAPH"] = "some_value"
    os.environ["ACCELERATE_DYNAMO_USE_DYNAMIC"] = "some_value"
    assert is_using_accelerate()


def test_is_using_accelerate_fsdp():
    if "ACCELERATE_USE_FSDP" in os.environ:
        del os.environ["ACCELERATE_USE_FSDP"]
    assert not is_using_accelerate_fsdp()

    os.environ["ACCELERATE_USE_FSDP"] = "false"
    assert not is_using_accelerate_fsdp()

    os.environ["ACCELERATE_USE_FSDP"] = "true"
    assert is_using_accelerate_fsdp()

    os.environ["ACCELERATE_USE_FSDP"] = "invalid_value"
    with pytest.raises(ValueError, match="Cannot convert 'invalid_value' to boolean."):
        is_using_accelerate_fsdp()



import datasets
import pandas as pd
import pytest

from oumi.core.datasets.base_map_dataset import BaseMapDataset


class MyTestDataset(BaseMapDataset):
    def __init__(self, *args, **kwargs):
        self._data = self._default_mock_data()
        super().__init__(*args, **kwargs)

    def _load_data(self):
        return self.mock_data

    def transform(self, sample: pd.Series) -> dict:
        result = sample.to_dict()
        computed_dict = {}
        for key, value in result.items():
            if isinstance(value, (str, list)):
                computed_dict[f"{key}_len"] = len(value)
        result.update(computed_dict)
        return result

    def _default_mock_data(self) -> pd.DataFrame:
        return pd.DataFrame(
            [
                {"text": "This is a test sentence."},
                {"text": "Another example sentence for testing."},
                {"text": "A third sentence to ensure we have enough data."},
            ]
        )


@pytest.mark.parametrize(
    "stream",
    [
        False,
        True,
    ],
)
def test_to_hf_basic(stream: bool):
    custom_dataset = MyTestDataset(dataset_name=f"foo_stream_{stream}")
    dataset = custom_dataset.to_hf(return_iterable=stream)

    if stream:
        assert isinstance(dataset, datasets.IterableDataset)
    else:
        assert isinstance(dataset, datasets.Dataset)
